unit BodaMaths;

interface
uses Dialogs,
     SysUtils,
     dglopengl;
 type
  Matrix3f = Array[0..2] of Array[0..2] of Extended;
  Vector3f = record
              x, y, z: Extended;
             end;
             

  
 function SubstractVectors(VectorA, VectorB: Vector3f): Vector3f;
 function CalcCrossProduct(VectorA, VectorB: Vector3f): Vector3f;
 function Normalize(Vector: Vector3f): Vector3f;
 function CalcNormalOfTriangle(Vertex1, Vertex2, Vertex3: Vector3f): Vector3f;
 function CalcTangentSpaceVectors(Vertex0, Vertex1, Vertex2, TexCoords0, TexCoords1, TexCoords2: Vector3f; Normal: Vector3f): Matrix3f;
 function MultVectors(VectorA, VectorB: Vector3f): Vector3f;
 function MultWithVector(VectorA: Vector3f; Zahl: Extended): Vector3f;
 function CalcDotProduct(VectorA, VectorB : Vector3f) : Extended;
 function CalcNegativeVector(Vector: Vector3f): Vector3f;
 procedure VectorMessage(theVector: Vector3f);

implementation

function CalcTangentSpaceVectors(Vertex0, Vertex1, Vertex2, TexCoords0, TexCoords1, TexCoords2: Vector3f; Normal: Vector3f): Matrix3f;
  var Vec1, Vec2: Vector3f;
      DeltaU1, DeltaU2, DeltaV1, DeltaV2: Extended;
      DirectionV, DirectionU, DirectionW: Vector3f;
      TempResult: Matrix3f;
      
begin
{ Vec1:= SubstractVectors(Vertex2, Vertex1);
 Vec2:= SubstractVectors(Vertex0, Vertex1);
 
 DeltaU1:= TexCoords0.x - TexCoords1.x;
 DeltaU2:= TexCoords2.x - TexCoords1.x;
 DirectionU:= SubstractVectors(MultWithVector(Vec1, DeltaU2), MultWithVector(Vec2, DeltaU1));
 DirectionU:= Normalize(DirectionU);

 DeltaV1:= TexCoords0.y - TexCoords1.y;
 DeltaV2:= TexCoords2.y - TexCoords1.y;
 DirectionV:= SubstractVectors(MultWithVector(Vec2, DeltaV1), MultWithVector(Vec1, DeltaV2));
 DirectionV:= Normalize(DirectionV);

 DirectionW:= CalcCrossProduct(DirectionV, DirectionU);
 DirectionW:= Normalize(DirectionW);     }


 { Vec1 = Vertex3  Vertex2
   Vec2 = Vertex1  Vertex2
   DeltaU1 = Vertex3.u - Vertex2.u
   DeltaU2 = Vertex1.u - Vertex2.u
   DirectionV = |DeltaU2*Vec1-DeltaU1*Vec2|
   DirectionU = |DirectionV x Vertex.N|
   DirectionW = |DirectionU x DirectionV|}

 Vec1:= SubstractVectors(Vertex2, Vertex1);
 Vec2:= SubstractVectors(Vertex0, Vertex1);

 DeltaU1:= TexCoords2.x - TexCoords1.x;
 DeltaU2:= TexCoords0.x - TexCoords1.x;

 DirectionV:= SubstractVectors(MultWithVector(Vec1, DeltaU2), MultWithVector(Vec2, DeltaU1));
 DirectionV:= normalize(DirectionV);


 DirectionU:= CalcCrossProduct(DirectionV, Normal);
 DirectionU:= normalize(DirectionU);


 DirectionW:= CalcCrossProduct(DirectionU, DirectionV);
 DirectionW:= normalize(DirectionW);    


 if (CalcDotProduct(DirectionW, Normal) < 0) then
 begin
  DirectionU:= CalcNegativeVector(DirectionU);
  DirectionV:= CalcNegativeVector(DirectionV);
  DirectionW:= CalcNegativeVector(DirectionW);
 end;

 TempResult[0][0]:= -DirectionU.x; TempResult[0][1]:= -DirectionU.y; TempResult[0][2]:= -DirectionU.z;
 TempResult[1][0]:= -DirectionV.x; TempResult[1][1]:= -DirectionV.y; TempResult[1][2]:= -DirectionV.z;
 TempResult[2][0]:= DirectionW.x; TempResult[2][1]:= DirectionW.y; TempResult[2][2]:= DirectionW.z;

 Result:= TempResult;
end;                                         

procedure VectorMessage(theVector: Vector3f);
 var tmpstr: String;
begin
 tmpstr:= '  ';
 tmpstr:= tmpstr + FloatToStr(theVector.x)+ '  ' + FloatToStr(theVector.y)+ '  ' + FloatToStr(theVector.z)+ '  ';
 ShowMessage(tmpstr);
end;

function CalcNegativeVector(Vector: Vector3f): Vector3f;
begin
 Result.x:= -Vector.x;
 Result.y:= -Vector.y;
 Result.z:= -Vector.z;
end;

function CalcNormalOfTriangle(Vertex1, Vertex2, Vertex3: Vector3f): Vector3f;
 var VectorA, VectorB, tmp: Vector3f;
begin
 VectorA:= SubstractVectors(Vertex1, Vertex2);
 VectorB:= SubstractVectors(Vertex2, Vertex3);
 tmp:= CalcCrossProduct(VectorA, VectorB);
 tmp:= Normalize(tmp);
 Result:= tmp;
end;

function CalcCrossProduct(VectorA, VectorB: Vector3f): Vector3f;
begin
 Result.x:= VectorA.y*VectorB.z - VectorA.z*VectorB.y;
 Result.y:= VectorA.z*VectorB.x - VectorA.x*VectorB.z;
 Result.z:= VectorA.x*VectorB.y - VectorA.y*VectorB.x;
end;

function Normalize(Vector: Vector3f): Vector3f;
 var
  Length: Extended;
begin
 Length:= sqrt(Vector.x*Vector.x + Vector.y*Vector.y + Vector.z*Vector.z);

 if Length <> 0 then
 begin
 Result.x:= Vector.x/Length;
 Result.y:= Vector.y/Length;
 Result.z:= Vector.z/Length;
 end;
end;

function SubstractVectors(VectorA, VectorB: Vector3f): Vector3f;
begin
 Result.x:= VectorA.x-VectorB.x;
 Result.y:= VectorA.y-VectorB.y;
 Result.z:= VectorA.z-VectorB.z;
end;

function MultVectors(VectorA, VectorB: Vector3f): Vector3f;
begin
 Result.x:= VectorA.x*VectorB.x;
 Result.y:= VectorA.y*VectorB.y;
 Result.z:= VectorA.z*VectorB.z;
end;

function MultWithVector(VectorA: Vector3f; Zahl: Extended): Vector3f;
begin
 Result.x:= VectorA.x*Zahl;
 Result.y:= VectorA.y*Zahl;
 Result.z:= VectorA.z*Zahl;
end;

function CalcDotProduct(VectorA, VectorB : Vector3f) : Extended;
begin
   Result := VectorA.x*VectorB.x+
             VectorA.y*VectorB.y+
             VectorA.z*VectorB.z;
end;

end.
