unit Figures;

interface
uses dglopengl, BodaMaths;

type
 Object3D = record
                 NumberOfVertices: Integer;
                 Vertices: Array of Vector3f;
                 TexCoords: Array of Vector3f;
                 Normals: Array of Vector3f;
                 TangentSpace: Array of Matrix3f;
                 Tangent, Binormal, Normal: Array of Vector3f;
                 DrawMode: GLuint;

                end;
 Object3DS = record
              Meshes: Array of Object3D;
              Textures, BTextures: Array of Gluint;
              HasTex, HasBTex: Array of Boolean;
              NumMeshes: Integer;
              MatID: Array of Integer;
             end;

function SetUpQuad: Object3D;
procedure DrawObject3D(theObject: Object3D);
procedure Draw3dsMesh(theObject: Object3DS);
function SetUpSphere(CX, CY, CZ, Radius : TGLFloat; N : Integer) : Object3D;
function SetUp3dsMesh(MeshFile, TexturePath: String): Object3DS;

implementation

uses OpenGL15_MainForm, SysUtils, Dialogs, gl3ds, textures;

procedure Draw3dsMesh(theObject: Object3DS);
var
   I, J, k: integer;
begin

for I := 0 to theObject.NumMeshes-1 do
begin
 glActiveTextureARB(GL_TEXTURE0_ARB);
 glBindTexture(GL_TEXTURE_2D, theObject.Textures[theObject.MatID[I]]);
 glActiveTextureARB(GL_TEXTURE1_ARB);
 glBindTexture(GL_TEXTURE_2D, theObject.BTextures[theObject.MatID[I]]);


glBegin(GL_TRIANGLES);
 for J := 0 to theObject.Meshes[I].NumberOfVertices-1 do
  begin
    with theObject.Meshes[I] do
    begin
     k:= j div 3;
     glMultiTexCoord3fARB(GL_TEXTURE0_ARB, TexCoords[j].x, TexCoords[j].y, TexCoords[j].z);
     glMultiTexCoord3fARB(GL_TEXTURE1_ARB, TexCoords[j].x, TexCoords[j].y, TexCoords[j].z);

     glMultiTexCoord3fARB(GL_TEXTURE2_ARB, Tangent[k].x,  Tangent[k].y,  Tangent[k].z);
     glMultiTexCoord3fARB(GL_TEXTURE3_ARB, Binormal[k].x, Binormal[k].y, Binormal[k].z);
     glMultiTexCoord3fARB(GL_TEXTURE4_ARB, Normal[k].x,   Normal[k].y,   Normal[k].z); 

     glNormal3f(Normals[k].x, Normals[k].y, Normals[k].z);
     glVertex3f(Vertices[J].x, Vertices[J].y, Vertices[J].z);
    end;
   end;

glEnd;

glUseProgramObjectARB(0);
glDisable(GL_LIGHTING);

glLineWidth(2);
glBegin(GL_LINES);
 for J := 0 to theObject.Meshes[I].NumberOfVertices-1  do
  begin
    with theObject.Meshes[I] do
    begin
      k:= j div 3;
   glColor3f(0, 1, 0);
       glVertex3f(Vertices[j].x, Vertices[j].y, Vertices[j].z);
       glVertex3f(Vertices[j].x + Normal[k].x*0.5, Vertices[j].y+ normal[k].y*0.5, Vertices[j].z+ normal[k].z*0.5);

  { glColor3f(1, 0, 0);
       glVertex3f(Vertices[j].x, Vertices[j].y, Vertices[j].z);
       glVertex3f(Vertices[j].x + tangent[k].x*0.5, Vertices[j].y+ tangent[k].y*0.5, Vertices[j].z+ tangent[k].z*0.5);

   glColor3f(0, 0, 1);
       glVertex3f(Vertices[j].x, Vertices[j].y, Vertices[j].z);
       glVertex3f(Vertices[j].x + binormal[k].x*0.5, Vertices[j].y+ binormal[k].y*0.5, Vertices[j].z+ binormal[k].z*0.5);  }
    end;
  end;
glEnd;
glDisable(GL_LIGHTING);
glUseProgramObjectARB(shader); 

end;

end;

function SetUp3dsMesh(MeshFile, TexturePath: String): Object3DS;
 var Mesh: TAll3dsMesh;
     I, J: integer;
     tmp_str: String;
     k: Integer;
     tmp: Integer;
begin
 Mesh:= TAll3dsMesh.Create(nil);
 Mesh.TexturePath:= TexturePath;
 Mesh.LoadFromFile(MeshFile);


 Result.NumMeshes:= Mesh.NumMeshes;
 SetLength(Result.Meshes, Result.NumMeshes);
 SetLength(Result.Textures, Mesh.NumMaterials);
 SetLength(Result.BTextures, Mesh.NumMaterials);
 SetLength(Result.HasTex, Mesh.NumMaterials);
 SetLength(Result.HasBTex, Mesh.NumMaterials);

 for I:= 0 to Mesh.NumMaterials-1 do
  begin
   LoadTexture( Mesh.Material[I].TextureFilename, Result.Textures[I], false);
   tmp_str:= copy(Mesh.Material[I].TextureFilename, 0, Length(Mesh.Material[I].TextureFilename) - 4) +'_nm' + copy(Mesh.Material[I].TextureFilename, length(Mesh.Material[I].TextureFilename) - 3, 4);
   if FileExists(tmp_str) then
   begin
    Result.HasBTex[I]:= true;
    LoadTexture(tmp_str, Result.BTextures[I], false);
   end
   else
    Result.HasBTex[I]:= false;
  end;

 for I:= 0 to Mesh.NumMeshes-1 do
 begin
  SetLength(Result.MatID, Mesh.NumMeshes);
  SetLength(Result.Meshes[I].Vertices, Mesh.Mesh[I].Fnumindices);
  SetLength(Result.Meshes[I].TexCoords, Mesh.Mesh[I].Fnumindices);
  SetLength(Result.Meshes[I].Normals, Mesh.Mesh[I].Fnumindices div 3 +1);
  SetLength(Result.Meshes[I].TangentSpace, Mesh.Mesh[I].Fnumindices div 3 +1);
  SetLength(Result.Meshes[I].Tangent, Mesh.Mesh[I].Fnumindices div 3 +1);
  SetLength(Result.Meshes[I].Binormal, Mesh.Mesh[I].Fnumindices div 3 +1);
  SetLength(Result.Meshes[I].Normal, Mesh.Mesh[I].Fnumindices div 3 +1);
  Result.Meshes[I].NumberOfVertices:= Mesh.Mesh[I].Fnumindices;


  with Result.Meshes[i] do
  begin

  for J:= 0 to Mesh.Mesh[I].Fnumindices-1 do
   begin
   Result.MatID[I]:= Mesh.Mesh[I].MatID[J div 3];


   Vertices[j].x:= Mesh.Mesh[I].Vertex[Mesh.Mesh[I].findices[J]].x;   //wird Vertex wirklich richtig ausgelesen
   Vertices[j].y:= Mesh.Mesh[I].Vertex[Mesh.Mesh[I].findices[J]].y;
   Vertices[j].z:= Mesh.Mesh[I].Vertex[Mesh.Mesh[I].findices[J]].z;

   TexCoords[j].x:= Mesh.Mesh[I].FMapping[Mesh.Mesh[I].findices[J]].tu;
   TexCoords[j].y:= Mesh.Mesh[I].FMapping[Mesh.Mesh[I].findices[J]].tv;
   TexCoords[j].z:= 0;

   if ((j+1) mod 3 = 0) then
     begin
      tmp:= j div 3;
      Normals[tmp]:=       CalcNormalOfTriangle(Vertices[j-2], Vertices[j-1], Vertices[j]);
      TangentSpace[tmp]:=  CalcTangentSpaceVectors(Vertices[j-2], Vertices[j-1], Vertices[j], texCoords[j-2], texCoords[j-1], texCoords[j], Normals[tmp]);
      Tangent[tmp].x:=     TangentSpace[tmp][0][0];
      Tangent[tmp].y:=     TangentSpace[tmp][0][1];
      Tangent[tmp].z:=     TangentSpace[tmp][0][2];
      Binormal[tmp].x:=    TangentSpace[tmp][1][0];
      Binormal[tmp].y:=    TangentSpace[tmp][1][1];
      Binormal[tmp].z:=    TangentSpace[tmp][1][2];
      Normal[tmp].x:=      TangentSpace[tmp][2][0];
      Normal[tmp].y:=      TangentSpace[tmp][2][1];
      Normal[tmp].z:=      TangentSpace[tmp][2][2];
     end;
  end;
 end;     

 end;
 Mesh.Destroy;
end;

procedure DestroyObject(theObject: Object3D);
begin
 with theObject do
 begin
  SetLength(Vertices, 1);
  SetLength(TexCoords, 1);
  SetLength(Normals, 1);
  SetLength(TangentSpace, 1);
  SetLength(Tangent, 1);
  SetLength(Binormal, 1);
  SetLength(Normal, 1);
 end;
end;

function SetUpQuad: Object3D;
 var Ergebnis: Object3D;
     i: Integer;
begin
 Ergebnis.DrawMode:= GL_QUADS;

with Ergebnis do
begin
 NumberOfVertices:= 4;
 SetLength(Normals, NumberOfVertices);
 SetLength(TexCoords, NumberOfVertices);
 SetLength(Vertices, NumberOfVertices);
 SetLength(Tangent, NumberOfVertices);
 SetLength(Binormal, NumberOfVertices);
 SetLength(Normal, NumberOfVertices);
 SetLength(TangentSpace, NumberOfVertices);

 
 TexCoords[0].x:= 0; TexCoords[0].y:= 0; TexCoords[0].z:= 0;
 Vertices[0].x:= 0; Vertices[0].y:= 0; Vertices[0].z:= 0;

 TexCoords[1].x:= 5; TexCoords[1].y:= 0; TexCoords[1].z:= 0;
 Vertices[1].x:= 5; Vertices[1].y:= 0; Vertices[1].z:= 0;

 TexCoords[2].x:= 5; TexCoords[2].y:= 5; TexCoords[2].z:= 0;
 Vertices[2].x:= 5; Vertices[2].y:= 5; Vertices[2].z:= 0;

 TexCoords[3].x:= 0; TexCoords[3].y:= 5; TexCoords[3].z:= 0;
 Vertices[3].x:= 0; Vertices[3].y:= 5; Vertices[3].z:= 0;

 Normals[0]:= CalcNormalOfTriangle(Vertices[0], Vertices[1], Vertices[2]);
 Normals[1]:= Normals[0];
 Normals[2]:= Normals[0];
 Normals[3]:= Normals[0];

 TangentSpace[0] := CalcTangentSpaceVectors(Vertices[0], Vertices[1], Vertices[2], TexCoords[0], TexCoords[1], TexCoords[2], Normals[0]);

 Tangent[0].x:= TangentSpace[0][0][0];
 Tangent[0].y:= TangentSpace[0][0][1];
 Tangent[0].z:= TangentSpace[0][0][2];
 Binormal[0].x:= TangentSpace[0][1][0];
 Binormal[0].y:= TangentSpace[0][1][1];
 Binormal[0].z:= TangentSpace[0][1][2];
 Normal[0].x:= TangentSpace[0][2][0];
 Normal[0].y:= TangentSpace[0][2][1];
 Normal[0].z:= TangentSpace[0][2][2];

end;
 Result:= Ergebnis;
end;

procedure DrawObject3D(theObject: Object3D);
 var i: Integer;
begin
 glBegin(theObject.DrawMode);

  for i:= 0 to theObject.NumberOfVertices-1 do
  begin
   glMultiTexCoord3fARB(GL_TEXTURE0_ARB, theObject.TexCoords[i].x, theObject.TexCoords[i].y, theObject.TexCoords[i].z);
   glMultiTexCoord3fARB(GL_TEXTURE1_ARB, theObject.TexCoords[i].x, theObject.TexCoords[i].y, theObject.TexCoords[i].z);
   glMultiTexCoord3fARB(GL_TEXTURE2_ARB, theObject.Tangent[0].x, theObject.Tangent[0].y, theObject.Tangent[0].z);
   glMultiTexCoord3fARB(GL_TEXTURE3_ARB, theObject.Binormal[0].x, theObject.Binormal[0].y, theObject.Binormal[0].z);
   glMultiTexCoord3fARB(GL_TEXTURE4_ARB, theObject.Normal[0].x, theObject.Normal[0].y, theObject.Normal[0].z);

   glNormal3f(theObject.Normals[i].x, theObject.Normals[i].y, theObject.Normals[i].z);
   glVertex3f(theObject.Vertices[i].x, theObject.Vertices[i].y, theObject.Vertices[i].z);
  end; 

 glEnd;

 with theObject do
 for i:= 0 to NumberOfVertices-1 do
  begin
  glUseProgramObjectARB(0);
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
     glColor3f(0, 1, 0);
       glVertex3f(Vertices[i].x, Vertices[i].y, Vertices[i].z);
       glVertex3f(Vertices[i].x + Normals[0].x, Vertices[i].y+ normals[0].y, Vertices[i].z+ normals[0].z);

    glColor3f(1, 0, 0);
       glVertex3f(Vertices[i].x, Vertices[i].y, Vertices[i].z);
       glVertex3f(Vertices[i].x + tangent[0].x, Vertices[i].y+ tangent[0].y, Vertices[i].z+ tangent[0].z);

    glColor3f(0, 0, 1);
       glVertex3f(Vertices[i].x, Vertices[i].y, Vertices[i].z);
       glVertex3f(Vertices[i].x + binormal[0].x, Vertices[i].y+ binormal[0].y, Vertices[i].z+ binormal[0].z);
  glEnd;
  glUseProgramObjectARB(shader);
 end;
end;

function SetUpSphere(CX, CY, CZ, Radius : TGLFloat; N : Integer) : Object3D;
var
 i,j                  : Integer;
 theta1,theta2,theta3 : TGLFloat;
 X, Y, Z, px, py, pz  : TGLFloat;

 index : Integer;
 ThreeVectors, ThreeTexCoords: Array[0..2] of Vector3f;
 TangentSpace: TGLMatrixf4;
begin
index:= 0;

if Radius < 0 then
 Radius :=-Radius;
if n < 0 then
 n := -n;
if (n < 4) or (Radius <= 0) then
 begin
 glBegin(GL_POINTS);
  glVertex3f(CX, CY, CZ);
 glEnd;
 exit;
 end;
for j := 0 to n div 2 -1 do
 begin
 theta1 := J*2*PI/N - PI/2;
 theta2 := (J+1)*2*PI/n - PI/2;
 glBegin(GL_TRIANGLE_STRIP);
 for i := 0 to n do
  begin
  
  theta3 := i*2*PI/N;
  x := cos(theta2) * cos(theta3);
  y := sin(theta2);
  z := cos(theta2) * sin(theta3);
  px := CX + Radius*x;
  py := CY + Radius*y;
  pz := CZ + Radius*z;
 
  glNormal3f(X,Y,Z);
  glTexCoord2f(1-I/n, 2*(J+1)/n);
  glVertex3f(px,py,pz);


  X := cos(theta1) * cos(theta3);
  Y := sin(theta1);
  Z := cos(theta1) * sin(theta3);
  px := CX + Radius*X;
  py := CY + Radius*Y;
  pz := CZ + Radius*Z;

  glNormal3f(X,Y,Z);
  glTexCoord2f(1-i/n, 2*j/n);
  glVertex3f(px,py,pz);

  end;
 glEnd;
 end;
end;

end.
