// =============================================================================
//   OpenGL1.5 - VCL Template (opengl15_vcl_template.zip)
// =============================================================================
//   Copyright  2003 by DGL - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
// =============================================================================
//   History :
//    Version 1.0 - Initial Release                            (Sascha Willems)
// =============================================================================

unit OpenGL15_MainForm;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,

  dglOpenGL,
  settings,
  keys,
  shader,
  textures,
  figures,
  glbitmap,
  StdCtrls, ExtCtrls, ComCtrls;

type
  TGLForm = class(TForm)
    Panel4: TPanel;
    Panel5: TPanel;
    ProgressBar1: TProgressBar;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  private
    { Private-Deklarationen }
  public
    RC        : HGLRC;
    DC        : HDC;
    ShowFPS   : Boolean;
    FontBase  : GLUInt;
    StartTick : Cardinal;
    Frames    : Integer;
    FPS       : Single;
  end;

var
  GLForm: TGLForm;
  SPEEDVARIABLE: Extended;

  shader: GLHandleARB;

  FCubeMap: TglBitmapCubeMap;

  Viereck: Object3D;
  Bump: Array[0..30] of GLuint;

  CounterBegin, CounterEnd, CounterFrequency: Int64;
  time: Extended;

implementation

{$R *.dfm}

procedure TGLForm.FormCreate(Sender: TObject);
 var i: Integer;
begin
 SettingsOnInit;

 shader:= LoadFragmentandVertexShader('fs.txt', 'vs.txt');


  FCubeMap := TglBitmapCubeMap.Create;

  FCubeMap.LoadFromFile('cubemap/norden.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_X);
  FCubeMap.SetTextureWrap(GL_REPEAT, GL_REPEAT, GL_REPEAT);
  FCubeMap.LoadFromFile('cubemap/sueden.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_X);
  FCubeMap.LoadFromFile('cubemap/alterna.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_Y);
  FCubeMap.LoadFromFile('cubemap/alterna.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y);
  FCubeMap.LoadFromFile('cubemap/westen.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_Z);
  FCubeMap.LoadFromFile('cubemap/osten.jpg');
  FCubeMap.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z);

  for i := 1 to 30 do
   LoadTexture(ExtractFilePath(Application.ExeName) +  '/nm/normalmap' + IntToStr(i) + '.jpg', Bump[i], false);

  QueryPerformanceFrequency(CounterFrequency);

  Viereck:= SetUpQuad;
end;
                                     
procedure TGLForm.FormDestroy(Sender: TObject);
begin
 SettingsOnDestroy;
end;

procedure TGLForm.ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
 var i: Integer;
     tmp: integer;
begin
 QueryPerformanceCounter(CounterBegin);
 SettingsOnIdle1;

 glUseProgramObjectARB(shader);

 glUniform1iARB(glGetUniformLocationARB(shader, 'Texture0'), 0);
 glUniform1iARB(glGetUniformLocationARB(shader, 'Texture1'), 1);

 glActiveTextureARB(GL_TEXTURE1_ARB);
 glEnable(GL_TEXTURE_2D);

 if time > 1000 then time:= time - 1000;
  tmp:= trunc(time/(1/30*1000));

 ProgressBar1.Position:= trunc(100/(30) * (tmp));

 glBindTexture(GL_TEXTURE_2D, Bump[1+tmp]);

 glActiveTextureARB(GL_TEXTURE0_ARB);
 glEnable(GL_TEXTURE_CUBE_MAP_ARB);

 glTranslatef(-15, -15, -30);
 glScalef(10, 10, 10);

 DrawObject3D(Viereck);

 SettingsOnIdle2;
 Done := False;                                  

 QueryPerformanceCounter(CounterEnd);
 if ((CounterEnd-CounterBegin)/CounterFrequency)*1000 < 100  then
  time:= time + ((CounterEnd-CounterBegin)/CounterFrequency)*1000;
end;

procedure TGLForm.FormKeyPress(Sender: TObject; var Key: Char);
begin

 processKeys(Key);
end;

end.
