(***********************************************************
glBitmap by Steffen Xonna (2003-2005)
http://www.dev-center.de/
------------------------------------------------------------
This unit implement some textureobjects wich have inspired by

glBMP.pas Copyright by Jason Allen
http://delphigl.cfxweb.net/

and

textures.pas Coypright by Jan Horn
http://www.sulaco.co.za/

It is fully compatible with an standard Delphi TBitmap.
------------------------------------------------------------
The contents of this file are used with permission, subject to
the Mozilla Public License Version 1.1 (the "License"); you may
not use this file except in compliance with the License. You may
obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html
------------------------------------------------------------
Version 1.3
------------------------------------------------------------
History
21-04-2005
- class TglBitmapCubeMap added (allows to Create Cubemaps)
29-03-2005
- Added Support for PNG Images. (http://pngdelphi.sourceforge.net/)
  To Enable png's use the define pngimage
22-03-2005
- New Functioninterface added
- Function GetPixel added
27-11-2004
- Property BuildMipMaps renamed to MipMap
21-11-2004
- property Name removed.
- BuildMipMaps is now a set of 3 values. None, GluBuildMipmaps and SGIS_generate_mipmap
22-05-2004
- property name added. Only used in glForms!
26-11-2003
- property FreeDataAfterGenTexture is now available as default (default = true)
- BuildMipmaps now implemented in TglBitmap1D (i've forgotten it)
- function MoveMemory replaced with function Move (little speed change)
- several calculations stored in variables (little speed change)
29-09-2003
- property BuildMipsMaps added (default = True)
  if BuildMipMaps isn't set GenTextures uses glTexImage[12]D else it use gluBuild[12]dMipmaps
- property FreeDataAfterGenTexture added (default = True)
  if FreeDataAfterGenTexture is set the texturedata were deleted after the texture was generated.
- parameter DisableOtherTextureUnits of Bind removed
- parameter FreeDataAfterGeneration of GenTextures removed
12-09-2003
- TglBitmap dosn't delete data if class was destroyed (fixed)
09-09-2003
- Bind now enables TextureUnits (by params)
- GenTextures can leave data (by param)
- LoadTextures now optimal
03-09-2003
- Performance optimization in AddFunc
- procedure Bind moved to subclasses 
- Added new Class TglBitmap1D to support real OpenGL 1D Textures
19-08-2003
- Texturefilter and texturewrap now also as defaults
  Minfilter = GL_LINEAR_MIPMAP_LINEAR
  Magfilter = GL_LINEAR
  Wrap(str) = GL_CLAMP_TO_EDGE
- Added new format tfCompressed to create a compressed texture.
- propertys IsCompressed, TextureSize and IsResident added
  IsCompressed and TextureSize only contains data from level 0
18-08-2003
- Added function AddFunc to add PerPixelEffects to Image
- LoadFromFunc now based on AddFunc
- Invert now based on AddFunc
- SwapColors now based on AddFunc
16-08-2003
- Added function FlipHorz
15-08-2003
- Added function LaodFromFunc to create images with function
- Added function FlipVert
- Added internal format RGB(A) if GL_EXT_bgra or OpenGL 1.2 isn't supported
29-07-2003
- Added Alphafunctions to calculate alpha per function
- Added Alpha from ColorKey using alphafunctions
28-07-2003
- First full functionally Version of glBitmap
- Support for 24Bit and 32Bit TGA Pictures added 
25-07-2003
- begin of programming
***********************************************************)
unit glBitmap;

interface

{$X+,H+}

{ $define pngimage}

uses
  Windows, SyncObjs, dglOpenGL, Graphics, Classes, SysUtils, JPEG{$ifdef pngimage}, pngimage{$endif};

type
  TglBitmap = class;

  // Exception
  EglBitmapException = Exception;
  EglBitmapSizeToLargeException = EglBitmapException;

  // Einstellungen
  TglBitmapTextureFormat = (tfDefault, tf16Bit, tf32Bit, tfCompressed);
  TglBitmapMipMap = (mmNone, mmMipmap, mmMipmapGlu);

  // Functions
  TglBitmapPixelDataFields = set of (ffRed, ffGreen, ffBlue, ffAlpha);
  TglBitmapPixelData = record
    Fields : TglBitmapPixelDataFields;
    ptRed : pByte;
    ptGreen : pByte;
    ptBlue : pByte;
    ptAlpha : pByte;
  end;

  TglBitmapPixelPositionFields = set of (ffX, ffY, ffZ);
  TglBitmapPixelPosition = record
    Fields : TglBitmapPixelPositionFields;
    X : Word;
    Y : Word;
    Z : Word;
  end;

  TglBitmapFunction = procedure(
    Sender : TglBitmap;
    const Position, Size: TglBitmapPixelPosition;
    const Source, Dest: TglBitmapPixelData;
    const Data: Pointer);

  TglBitmapGetPixel = procedure (
    const Pos: TglBitmapPixelPosition;
    var Pixel: TglBitmapPixelData) of object;


  // Base Class
  TglBitmap = class
  protected
    FTextureID: TGLuint;
    FTextureTarget: Cardinal;
    FTextureFormat: TglBitmapTextureFormat;
    FMipMap: TglBitmapMipMap;
    FDeleteTextureOnFree: Boolean;
    FFreeDataAfterGenTexture: Boolean;

    // Propertys
    FDataPtr: PByte;
    FHasAlpha: Boolean;
    FIsCompressed: Boolean;
    FTextureSize: Integer;
    FIsResident: Boolean;

    // Filtering
    FFilterMin: TGLint;
    FFilterMag: TGLint;

    // Texturwarp
    FWrapS: TGLint;
    FWrapT: TGLint;
    FWrapR: TGLint;

    FGetPixelFunc: TglBitmapGetPixel;

    function GetData: PByte;  
    procedure AllocData(Size: Integer);
    procedure SetDataPtr(Ptr: PByte); virtual;

    procedure SwapColors; virtual;

    {$ifdef pngimage}
    function LoadPng(const Stream: TStream): Boolean; virtual;
    {$endif}
    function LoadTga(const Stream: TStream): Boolean; virtual;
    function LoadJpg(const Stream: TStream): Boolean; virtual;
    function LoadBmp(const Stream: TStream): Boolean; virtual;

    procedure CreateTextureID;
    procedure SetupTextureParameters(var BuildWithGlu: Boolean);
    procedure SelectTextureFormat(var Format, Components: Cardinal);
  public
    // propertys
    property TextureID: TGLuint read FTextureID;

    property TextureTarget: Cardinal read FTextureTarget
      write FTextureTarget;

    property TextureFormat: TglBitmapTextureFormat read FTextureFormat
      write FTextureFormat;

    property MipMap: TglBitmapMipMap read FMipMap write FMipMap;

    property DeleteTextureOnFree: Boolean read FDeleteTextureOnFree
      write FDeleteTextureOnFree;

    property FreeDataAfterGenTexture: Boolean read FFreeDataAfterGenTexture
      write FFreeDataAfterGenTexture;

    property HasAlpha: Boolean read FHasAlpha;
    property IsCompressed: Boolean read FIsCompressed;
    property TextureSize: Integer read FTextureSize;
    property IsResident: boolean read FIsResident;

    // Construction and Destructions Methods
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;

    constructor Create(); overload;
    constructor Create(FileName: String); overload;
    constructor CreateFromResourceName(Name: String); 
    constructor Create(Stream: TStream); overload;
    constructor Create(ResourceID: Integer); overload;

    // Loading Methods
    procedure LoadFromFile(FileName: String);
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure LoadFromResource(Resource: String);
    procedure LoadFromResourceID(ResourceID: Integer);

    function AddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer = nil): boolean; virtual; abstract;

    function AssignToBitmap(const Bitmap: TBitmap): boolean; virtual; abstract;
    function AssignAlphaToBitmap(const Bitmap: TBitmap): boolean; virtual; abstract;
    function AssignFromBitmap(const Bitmap: TBitmap): boolean; virtual; abstract;

    function AddAlphaFromFunc(Func: TglBitmapFunction; Data: Pointer = nil): boolean; virtual; abstract;
    function AddAlphaFromBitmap(Bitmap: TBitmap; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean; virtual; abstract;
    function AddAlphaFromFile(FileName: String; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean;
    function AddAlphaFromStream(Stream: TStream; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean;
    function AddAlphaFromResource(Resource: String; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean;
    function AddAlphaFromResourceID(ResourceID: Integer; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean;

    function AddAlphaFromColorKey(Red, Green, Blue: Byte; Deviation: Byte = 0): Boolean;

    function RemoveAlpha: Boolean; virtual; abstract;

    // Other
    procedure Invert(UseRGB: Boolean = true; UseAlpha: Boolean = false); 
    procedure SetFilter(Min, Mag : TGLint);
    procedure SetTextureWrap(S: TGLint = GL_CLAMP_TO_EDGE;
      T: TGLint = GL_CLAMP_TO_EDGE; R: TGLint = GL_CLAMP_TO_EDGE);

    procedure GetPixel (const Pos: TglBitmapPixelPosition; var Pixel: TglBitmapPixelData); virtual; 

    // Generation
    procedure GenTexture(TestTextureSize: Boolean = True); virtual; abstract;
    procedure Bind(EnableTextureUnit: Boolean = True); virtual;
  end;


  TglBitmap2D = class(TglBitmap)
  protected
    // Bildeinstellungen
    FWidth: Integer;
    FHeight: Integer;
    FLines: array of PByte;

    procedure GetPixel2D (const Pos: TglBitmapPixelPosition; var Pixel: TglBitmapPixelData);
    procedure GetPixel2DAlpha (const Pos: TglBitmapPixelPosition; var Pixel: TglBitmapPixelData);

    function IntAddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean; 
    function IntAddFuncAlpha(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean; 

    procedure SetDataPtr(Ptr: PByte); override;

    procedure UploadTextureData (TextureTarget, Format, Components: Cardinal; BuildWithGlu: Boolean);
  public
    // propertys
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;

    procedure AfterConstruction; override;

    constructor Create(Width, Height: Integer; Func: TglBitmapFunction; HasAlpha: Boolean); overload;

    // Loading Methods
    procedure LoadFromFunc(Width, Height: Integer; Func: TglBitmapFunction; HasAlpha: Boolean);

    function AddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer = nil): boolean; override;

    function AssignToBitmap(const Bitmap: TBitmap): boolean; override;
    function AssignAlphaToBitmap(const Bitmap: TBitmap): boolean; override;
    function AssignFromBitmap(const Bitmap: TBitmap): boolean; override;

    function AddAlphaFromFunc(Func: TglBitmapFunction; Data: Pointer = nil): boolean; override;
    function AddAlphaFromBitmap(Bitmap: TBitmap; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean; override;

    function RemoveAlpha: Boolean; override;

    // Other
    function FlipHorz: Boolean;
    function FlipVert: Boolean; 

    // Generation
    procedure GenTexture(TestTextureSize: Boolean = True); override;
  end;


  TglBitmapCubeMap = class(TglBitmap2d)
  protected
    // Hide GenTexture
    procedure GenTexture(TestTextureSize: Boolean = True); reintroduce;
  public
    procedure AfterConstruction; override;

    procedure GenerateCubeMap(CubeTarget: Cardinal);

    procedure Bind(EnableTexCoordsGen: Boolean = true; EnableTextureUnit: Boolean = True); reintroduce; virtual; 
  end;


  TglBitmap1D = class(TglBitmap)
  protected
    // Bildeinstellungen
    FWidth: Integer;

    procedure GetPixel1D (const Pos: TglBitmapPixelPosition; var Pixel: TglBitmapPixelData);
    procedure GetPixel1DAlpha (const Pos: TglBitmapPixelPosition; var Pixel: TglBitmapPixelData);

    function IntAddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
    function IntAddFuncAlpha(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;

    procedure SetDataPtr(Ptr: PByte); override;

    procedure UploadTextureData (TextureTarget, Format, Components: Cardinal; BuildWithGlu: Boolean);
  public
    // propertys
    property Width: Integer read FWidth;

    procedure AfterConstruction; override;

    constructor Create(Width: Integer; Func: TglBitmapFunction; HasAlpha: Boolean); overload;

    // Loading Methods
    procedure LoadFromFunc(Width: Integer; Func: TglBitmapFunction; HasAlpha: Boolean);

    function AddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer = nil): boolean; override;

    function AssignToBitmap(const Bitmap: TBitmap): boolean; override;
    function AssignAlphaToBitmap(const Bitmap: TBitmap): boolean; override;
    function AssignFromBitmap(const Bitmap: TBitmap): boolean; override;

    function AddAlphaFromFunc(Func: TglBitmapFunction; Data: Pointer = nil): boolean; override;
    function AddAlphaFromBitmap(Bitmap: TBitmap; Func: TglBitmapFunction = nil; Data: Pointer = nil): boolean; override;

    function RemoveAlpha: Boolean; override;

    // Other
    function FlipHorz: Boolean;

    // Generation
    procedure GenTexture(TestTextureSize: Boolean = True); override;
  end;


// methods and vars for Defaults
procedure glBitmapSetDefaultTextureFormat(TextureFormat: TglBitmapTextureFormat);
procedure glBitmapSetDefaultDeleteTextureOnFree(DeleteTextureOnFree: Boolean);
procedure glBitmapSetDefaultFilter(Min, Mag: TGLint);
procedure glBitmapSetDefaultTextureWrap(S: TGLint = GL_CLAMP_TO_EDGE;
  T: TGLint = GL_CLAMP_TO_EDGE; R: TGLint = GL_CLAMP_TO_EDGE);
procedure glBitmapSetDefaultFreeDataAfterGenTexture(FreeData: Boolean);

function glBitmapGetDefaultTextureFormat: TglBitmapTextureFormat;
function glBitmapGetDefaultDeleteTextureOnFree: Boolean;
procedure glBitmapGetDefaultFilter(var Min, Mag: TGLint);
procedure glBitmapGetDefaultTextureWrap(var S, T, R: TGLint);
function glBitmapGetDefaultFreeDataAfterGenTexture: Boolean;

// Call LoadingMethods
function LoadTexture(Filename: String; var Texture: TGLuint; LoadFromRes : Boolean): Boolean;

function LoadCubeMap(PositiveX, NegativeX, PositiveY, NegativeY, PositiveZ, NegativeZ: String; LoadFromRes : Boolean): TGLuint;
function LoadCubeMapClass(PositiveX, NegativeX, PositiveY, NegativeY, PositiveZ, NegativeZ: String; LoadFromRes : Boolean): TglBitmapCubeMap; 


var
  glBitmapTextureFormat: TglBitmapTextureFormat;
  glBitmapDeleteTextureOnFree: Boolean;
  glBitmapDefaultFilterMin: TGLint;
  glBitmapDefaultFilterMag: TGLint;
  glBitmapDefaultWrapS: TGLint;
  glBitmapDefaultWrapT: TGLint;
  glBitmapDefaultWrapR: TGLint;
  glBitmapFreeDataAfterGenTextures: Boolean;


implementation


uses Math;


type
  TTGAHeader = packed record
    aType: Byte;
    ColorMapType: Byte;
    ImageType: Byte;
    ColorMapSpec: Array[0..4] of Byte;
    OrigX: Word;
    OrigY: Word;
    Width: Word;
    Height: Word;
    Bpp: Byte;
    ImageDes: Byte;
  end;


procedure glBitmapDefaultAlphaFunc(Sender : TglBitmap; const Position, Size: TglBitmapPixelPosition; const Source, Dest: TglBitmapPixelData; const Data: Pointer);
begin
  Dest.ptAlpha^ := (Source.ptRed^ + Source.ptGreen^ + Source.ptBlue^) div 3;
end;


// Helper functions
function LoadTexture(Filename: String; var Texture: TGLuint; LoadFromRes : Boolean): Boolean;
var
  glBitmap: TglBitmap2D;
begin
  Result := false;
  Texture := 0;

  if (LoadFromRes)
    then glBitmap := TglBitmap2D.CreateFromResourceName(FileName)
    else glBitmap := TglBitmap2D.Create(FileName);

  try
    glBitmap.DeleteTextureOnFree := False;
    glBitmap.FreeDataAfterGenTexture := False;
    glBitmap.GenTexture(True);
    if (glBitmap.TextureID > 0) then begin
      Texture := glBitmap.TextureID;
      Result := True;
    end;
  finally
    glBitmap.Free;
  end;
end;


function LoadCubeMap(PositiveX, NegativeX, PositiveY, NegativeY, PositiveZ, NegativeZ: String; LoadFromRes : Boolean): TGLuint;
var
  CM: TglBitmapCubeMap;
begin
  Result := 0;
  CM := nil;

  try
    CM := LoadCubeMapClass(PositiveX, NegativeX, PositiveY, NegativeY, PositiveZ, NegativeZ, LoadFromRes);

    if Assigned (CM) then
      begin
        CM.DeleteTextureOnFree := False;

        Result := CM.TextureID;
      end;
  finally
    CM.Free;
  end;
end;


function LoadCubeMapClass(PositiveX, NegativeX, PositiveY, NegativeY, PositiveZ, NegativeZ: String; LoadFromRes : Boolean): TglBitmapCubeMap;
begin
  Result := TglBitmapCubeMap.Create;
  try
    // Settings
    Result.FreeDataAfterGenTexture := False;

    // maps
    if (LoadFromRes)
      then Result.LoadFromResource(PositiveX)
      else Result.LoadFromFile(PositiveX);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_X);

    if (LoadFromRes)
      then Result.LoadFromResource(NegativeX)
      else Result.LoadFromFile(NegativeX);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_X);

    if (LoadFromRes)
      then Result.LoadFromResource(PositiveY)
      else Result.LoadFromFile(PositiveY);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_Y);

    if (LoadFromRes)
      then Result.LoadFromResource(NegativeY)
      else Result.LoadFromFile(NegativeY);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y);

    if (LoadFromRes)
      then Result.LoadFromResource(PositiveZ)
      else Result.LoadFromFile(PositiveZ);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_POSITIVE_Z);

    if (LoadFromRes)
      then Result.LoadFromResource(NegativeZ)
      else Result.LoadFromFile(NegativeZ);
    Result.GenerateCubeMap(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z);
  except
    Result.Free;
    Result := nil;
  end;
end;


// Defaults
procedure glBitmapSetDefaultTextureFormat(TextureFormat: TglBitmapTextureFormat);
begin
  glBitmapTextureFormat := TextureFormat;
end;


procedure glBitmapSetDefaultDeleteTextureOnFree(DeleteTextureOnFree: Boolean);
begin
  glBitmapDeleteTextureOnFree := DeleteTextureOnFree;
end;


procedure glBitmapSetDefaultFilter(Min, Mag: TGLint);
begin
  case min of
    GL_NEAREST:
      glBitmapDefaultFilterMin := GL_NEAREST;
    GL_LINEAR:
      glBitmapDefaultFilterMin := GL_LINEAR;
    GL_NEAREST_MIPMAP_NEAREST:
      glBitmapDefaultFilterMin := GL_NEAREST_MIPMAP_NEAREST;
    GL_LINEAR_MIPMAP_NEAREST:
      glBitmapDefaultFilterMin := GL_LINEAR_MIPMAP_NEAREST;
    GL_NEAREST_MIPMAP_LINEAR:
      glBitmapDefaultFilterMin := GL_NEAREST_MIPMAP_LINEAR;
    GL_LINEAR_MIPMAP_LINEAR:
      glBitmapDefaultFilterMin := GL_LINEAR_MIPMAP_LINEAR;
  else
    raise EglBitmapException.Create('glBitmapSetDefaultFilter - Unknow Minfilter.');
  end;

  case mag of
    GL_NEAREST:
      glBitmapDefaultFilterMag := GL_NEAREST;
    GL_LINEAR:
      glBitmapDefaultFilterMag := GL_LINEAR;
  else
    raise EglBitmapException.Create('glBitmapSetDefaultFilter - Unknow Magfilter.');
  end;
end;


procedure glBitmapSetDefaultTextureWrap(S: TGLint; T: TGLint; R: TGLint);
begin
  case S of
    GL_CLAMP:
      glBitmapDefaultWrapS := GL_CLAMP;
    GL_REPEAT:
      glBitmapDefaultWrapS := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      glBitmapDefaultWrapS := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      glBitmapDefaultWrapS := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('glBitmapSetDefaultTextureWrap - Unknow Texturewrap(s).');
  end;

  case T of
    GL_CLAMP:
      glBitmapDefaultWrapT := GL_CLAMP;
    GL_REPEAT:
      glBitmapDefaultWrapT := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      glBitmapDefaultWrapT := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      glBitmapDefaultWrapT := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('glBitmapSetDefaultTextureWrap - Unknow Texturewrap(t).');
  end;

  case R of
    GL_CLAMP:
      glBitmapDefaultWrapR := GL_CLAMP;
    GL_REPEAT:
      glBitmapDefaultWrapR := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      glBitmapDefaultWrapR := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      glBitmapDefaultWrapR := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('glBitmapSetDefaultTextureWrap - Unknow Texturewrap(r).');
  end;
end;


procedure glBitmapSetDefaultFreeDataAfterGenTexture(FreeData: Boolean);
begin
  glBitmapFreeDataAfterGenTextures := FreeData;
end;


function glBitmapGetDefaultTextureFormat: TglBitmapTextureFormat;
begin
  Result := glBitmapTextureFormat;
end;


function glBitmapGetDefaultDeleteTextureOnFree: Boolean;
begin
  Result := glBitmapDeleteTextureOnFree;
end;


procedure glBitmapGetDefaultFilter(var Min, Mag: TGLint);
begin
  Min := glBitmapDefaultFilterMin;
  Mag := glBitmapDefaultFilterMag;
end;


procedure glBitmapGetDefaultTextureWrap(var S, T, R: TGLint);
begin
  S := glBitmapDefaultWrapS;
  T := glBitmapDefaultWrapT;
  R := glBitmapDefaultWrapR;
end;


function glBitmapGetDefaultFreeDataAfterGenTexture: Boolean;
begin
  Result := glBitmapFreeDataAfterGenTextures;
end;


{ TglBitmap }

procedure TglBitmap.AfterConstruction;
begin
  inherited;

  FTextureID := 0;
  FTextureTarget := 0;
  FMipMap := mmMipmap;
  FIsCompressed := False;
  FTextureSize := 0;
  FIsResident := False;

  // get defaults
  FFreeDataAfterGenTexture := glBitmapGetDefaultFreeDataAfterGenTexture;
  FDeleteTextureOnFree := glBitmapGetDefaultDeleteTextureOnFree;
  FTextureFormat := glBitmapGetDefaultTextureFormat;
  glBitmapGetDefaultFilter(FFilterMin, FFilterMag);
  glBitmapGetDefaultTextureWrap(FWrapS, FWrapT, FWrapR);
end;


procedure TglBitmap.BeforeDestruction;
begin
  SetDataPtr(nil);
  
  if ((FTextureID > 0) and (FDeleteTextureOnFree))
    then glDeleteTextures(1, @FTextureID);

  inherited;
end;


constructor TglBitmap.Create;
begin
  Assert(ClassType <> TglBitmap, 'Don''t create TglBitmap');

  inherited Create;
end;


constructor TglBitmap.Create(FileName: String);
begin
  Create;
  LoadFromFile(FileName);
end;


constructor TglBitmap.CreateFromResourceName(Name: String);
begin
  Create;
  LoadFromResource(Name)
end;


constructor TglBitmap.Create(Stream: TStream);
begin
  Create;
  LoadFromStream(Stream);
end;


constructor TglBitmap.Create(ResourceID: Integer);
begin
  Create;
  LoadFromResourceID(ResourceID);
end;


procedure TglBitmap.LoadFromFile(FileName: String);
var
  FS: TFileStream;
begin
  FS := TFileStream.Create(FileName, fmOpenRead);
  try
    FS.Position := 0;
    LoadFromStream(FS);
  finally
    FS.Free;
  end;
end;


procedure TglBitmap.LoadFromStream(Stream: TStream);
begin
  {$ifdef pngimage}
  if (not LoadPng(Stream))
    then
  {$endif}
  if (not LoadTga(Stream))
    then
  if (not LoadJpg(Stream))
    then
  if (not LoadBmp(Stream))
    then raise EglBitmapException.Create('TglBitmap.LoadFromStream - Couldn''t load Stream. It''s possible to be an unknow Streamtype.');
end;


procedure TglBitmap.LoadFromResource(Resource: String);
var
  RS: TResourceStream;
begin
  RS := TResourceStream.Create(HInstance, Resource, RT_BITMAP);
  if (not assigned(RS))
    then RS := TResourceStream.Create(HInstance, Resource, RT_RCDATA);

  if (not assigned(RS))
    then raise EglBitmapException.Create('TglBitmap.LoadFromResource - Resourcename were not found.');

  try
    LoadFromStream(RS);
  finally
    RS.Free;
  end;
end;


procedure TglBitmap.LoadFromResourceID(ResourceID: Integer);
var
  RS: TResourceStream;
begin
  RS := TResourceStream.CreateFromID(HInstance, ResourceID, RT_BITMAP);
  if (not assigned(RS))
    then RS := TResourceStream.CreateFromID(HInstance, ResourceID, RT_RCDATA);

  if (not assigned(RS))
    then raise EglBitmapException.Create('TglBitmap.LoadFromResourceID - ResourceID were not found.');

  try
    LoadFromStream(RS);
  finally
    RS.Free;
  end;
end;


function TglBitmap.AddAlphaFromFile(FileName: String;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  FS: TFileStream;
begin
  FS := TFileStream.Create(FileName, fmOpenRead);
  try
    Result := AddAlphaFromStream(FS, Func, Data);
  finally
    FS.Free;
  end;
end;


function TglBitmap.AddAlphaFromStream(Stream: TStream;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  bmp: TBitmap;
  glBitmap: TglBitmap2D;
begin
  assert(Assigned(GetData()), 'TglBitmap.AddAlphaFromStream - AddAlpha can only called if data where loaded.');

  glBitmap := TglBitmap2D.Create(Stream);
  try
    bmp := TBitmap.Create;
    try
      glBitmap.AssignToBitmap(bmp);
      bmp.PixelFormat := pf24bit;
      Result := AddAlphaFromBitmap(bmp, Func, Data);
    finally
      bmp.Free;
    end;
  finally
    glBitmap.Free;
  end;
end;


function TglBitmap.AddAlphaFromResource(Resource: String;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  RS: TResourceStream;
begin
  RS := TResourceStream.Create(HInstance, Resource, RT_BITMAP);
  if (not assigned(RS))
    then RS := TResourceStream.Create(HInstance, Resource, RT_RCDATA);

  if (not assigned(RS))
    then raise EglBitmapException.Create('TglBitmap.AddAlphaFromResource - Resourcename were not found.');

  try
    Result := AddAlphaFromStream(RS, Func, Data);
  finally
    RS.Free;
  end;
end;


function TglBitmap.AddAlphaFromResourceID(ResourceID: Integer;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  RS: TResourceStream;
begin
  RS := TResourceStream.CreateFromID(HInstance, ResourceID, RT_BITMAP);
  if (not assigned(RS))
    then RS := TResourceStream.CreateFromID(HInstance, ResourceID, RT_RCDATA);

  if (not assigned(RS))
    then raise EglBitmapException.Create('TglBitmap.AddAlphaFromResource - Resourcename were not found.');

  try
    Result := AddAlphaFromStream(RS, Func, Data);
  finally
    RS.Free;
  end;
end;


procedure glBitmapColorKeyAlphaFunc(Sender : TglBitmap; const Position, Size: TglBitmapPixelPosition; const Source, Dest: TglBitmapPixelData; const Data: Pointer);
begin
  with PRGBQuad(Data)^ do
    if ((Dest.ptRed^   <= rgbRed   + rgbReserved) and (Dest.ptRed^   >= rgbRed   - rgbReserved) and
        (Dest.ptGreen^ <= rgbGreen + rgbReserved) and (Dest.ptGreen^ >= rgbGreen - rgbReserved) and
        (Dest.ptBlue^  <= rgbRed   + rgbReserved) and (Dest.ptBlue^  >= rgbBlue  - rgbReserved))
      then Dest.ptAlpha^ := 0
      else Dest.ptAlpha^ := 255;
end;


function TglBitmap.AddAlphaFromColorKey(Red, Green, Blue, Deviation: Byte): Boolean;
var
  Data: TRGBQuad;
begin
  Data.rgbRed := Red;
  Data.rgbGreen := Green;
  Data.rgbBlue := Blue;
  Data.rgbReserved := Deviation;

  Result := AddAlphaFromFunc(glBitmapColorKeyAlphaFunc, @Data);
end;


procedure glBitmapInvertFunc(Sender : TglBitmap; const Position, Size: TglBitmapPixelPosition; const Source, Dest: TglBitmapPixelData; const Data: Pointer);
begin
  if (Integer(Data) and $1 > 0) then begin
    Dest.ptRed^   := 255 - Dest.ptRed^;
    Dest.ptGreen^ := 255 - Dest.ptGreen^;
    Dest.ptBlue^  := 255 - Dest.ptBlue^;
  end;

  if (Integer(Data) and $2 > 0) then begin
    Dest.ptAlpha^ := 255 - Dest.ptAlpha^;
  end;
end;


procedure TglBitmap.Invert(UseRGB, UseAlpha: Boolean);
begin
  if ((UseRGB) or (UseAlpha)) then begin
    AddFunc(glBitmapInvertFunc, False, Pointer(Integer(UseAlpha and HasAlpha) shl 1 or Integer(UseRGB)));
  end;
end;


procedure TglBitmap.SetFilter(Min, Mag: TGLint);
begin
  case Min of
    GL_NEAREST:
      FFilterMin := GL_NEAREST;
    GL_LINEAR:
      FFilterMin := GL_LINEAR;
    GL_NEAREST_MIPMAP_NEAREST:
      FFilterMin := GL_NEAREST_MIPMAP_NEAREST;
    GL_LINEAR_MIPMAP_NEAREST:
      FFilterMin := GL_LINEAR_MIPMAP_NEAREST;
    GL_NEAREST_MIPMAP_LINEAR:
      FFilterMin := GL_NEAREST_MIPMAP_LINEAR;
    GL_LINEAR_MIPMAP_LINEAR:
      FFilterMin := GL_LINEAR_MIPMAP_LINEAR;
  else
    raise EglBitmapException.Create('TglBitmap.SetFilter - Unknow Minfilter.');
  end;

  case Mag of
    GL_NEAREST:
      FFilterMag := GL_NEAREST;
    GL_LINEAR:
      FFilterMag := GL_LINEAR;
  else
    raise EglBitmapException.Create('TglBitmap.SetFilter - Unknow Magfilter.');
  end;
end;


procedure TglBitmap.SetTextureWrap(S: TGLint; T: TGLint; R: TGLint);
begin
  case S of
    GL_CLAMP:
      FWrapS := GL_CLAMP;
    GL_REPEAT:
      FWrapS := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      FWrapS := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      FWrapS := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('TglBitmap.SetTextureWrap - Unknow Texturewrap(s).');
  end;

  case T of
    GL_CLAMP:
      FWrapT := GL_CLAMP;
    GL_REPEAT:
      FWrapT := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      FWrapT := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      FWrapT := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('TglBitmap.SetTextureWrap - Unknow Texturewrap(t).');
  end;

  case R of
    GL_CLAMP:
      FWrapR := GL_CLAMP;
    GL_REPEAT:
      FWrapR := GL_REPEAT;
    GL_CLAMP_TO_EDGE:
      FWrapR := GL_CLAMP_TO_EDGE;
    GL_CLAMP_TO_BORDER:
      FWrapR := GL_CLAMP_TO_BORDER;
  else
    raise EglBitmapException.Create('TglBitmap.SetTextureWrap - Unknow Texturewrap(r).');
  end;
end;


function TglBitmap.GetData: PByte;
begin
  Result := FDataPtr;
end;


procedure TglBitmap.AllocData(Size: Integer);
begin
  SetDataPtr(AllocMem(Size));
end;


procedure TglBitmap.SetDataPtr(Ptr: PByte);
begin
  if FDataPtr <> Ptr then begin
    if (Assigned(FDataPtr))
      then FreeMem(FDataPtr);

    FDataPtr := Ptr;
  end;
end;


function TglBitmap.LoadBmp(const Stream: TStream): Boolean;
var
  bmp: TBitmap;
  StreamPos: Int64;
  Temp: array[0..1]of char;
begin
  Result := False;

  // reading first two bytes to test file and set cursor back to begin
  StreamPos := Stream.Position;
  Stream.Read(Temp[0], 2);
  Stream.Position := StreamPos;

  // if Bitmap then read file.
  if ((Temp[0] = 'B') and (Temp[1] = 'M')) then begin
    bmp := TBitmap.Create;
    try
      bmp.LoadFromStream(Stream);
      if bmp.PixelFormat in [pfDevice, pf1Bit, pf4Bit, pf8Bit, pf15Bit, pf16Bit, pfCustom]
        then bmp.PixelFormat := pf24bit;
      Result := AssignFromBitmap(bmp);
    finally
      bmp.Free;
    end;
  end;
end;


function TglBitmap.LoadJpg(const Stream: TStream): Boolean;
var
  bmp: TBitmap;
  jpg: TJPEGImage;
  StreamPos: Int64;
  Temp: array[0..1]of char;
begin
  Result := False;

  // reading first two bytes to test file and set cursor back to begin
  StreamPos := Stream.Position;
  Stream.Read(Temp[0], 2);
  Stream.Position := StreamPos;

  // if Bitmap then read file.
  if ((Temp[0] = chr($FF)) and (Temp[1] = chr($D8))) then begin
    bmp := TBitmap.Create;
    try
      jpg := TJPEGImage.Create;
      try
        jpg.LoadFromStream(Stream);
        bmp.Assign(jpg);
        Result := AssignFromBitmap(bmp);
      finally
        jpg.Free;
      end;
    finally
      bmp.Free;
    end;
  end;
end;


function TglBitmap.LoadTga(const Stream: TStream): Boolean;
var
  Header: TTGAHeader;
  bmp: TBitmap;
  pData: PByte;
  StreamPos: Int64;
  Row, RowSize: Integer;
begin
  Result := False;

  // reading header to test file and set cursor back to begin
  StreamPos := Stream.Position;
  Stream.Read(Header, SizeOf(Header));
  Stream.Position := StreamPos;

  // uncompressed images
  if (Header.ImageType = 2) then begin
    // no colormapped files
    if (Header.ColorMapType = 0) then begin
      if ((Header.Bpp = 32) or (Header.Bpp = 24)) then begin
        bmp := TBitmap.Create;
        try
          bmp.Width := Header.Width;
          bmp.Height := Header.Height;
          if (Header.Bpp = 24) then begin
            bmp.PixelFormat := pf24bit;
            FHasAlpha := False;
          end else begin
            bmp.PixelFormat := pf32bit;
            FHasAlpha := True;
          end;
          Stream.Position := StreamPos + SizeOf(Header);

          RowSize := Header.Width * (Header.Bpp div 8);

          for Row := 0 to Header.Height -1 do begin
            pData := bmp.Scanline[Header.Height - Row - 1];

            if (Assigned(pData))
              then Stream.Read(pData^, RowSize);
          end;

          Result := AssignFromBitmap(bmp);
        finally
          bmp.Free;
        end;
      end;
    end;
  end;
end;


{$ifdef pngimage}
function TglBitmap.LoadPng(const Stream: TStream): Boolean;
var
  StreamPos: Int64;
  Png: TPNGObject;
  Bmp: TBitmap;
  Header: Array[0..7] of Char;
  Row, Col, LineSize: Integer;
  pSource, pDest: pByte;

const
  PngHeader: Array[0..7] of Char = (#137, #80, #78, #71, #13, #10, #26, #10);

begin
  Result := False;

  StreamPos := Stream.Position;
  Stream.Read(Header[0], SizeOf(Header));
  Stream.Position := StreamPos;

  {Test if the header matches}
  if Header = PngHeader then begin
    Png := TPNGObject.Create;
    try
      Png.LoadFromStream(Stream);

      Bmp := TBitmap.Create;
      try
        Bmp.Width := Png.Width;
        Bmp.Height := Png.Height;
        case Png.Header.ColorType of
          COLOR_RGB:
            begin
              Bmp.PixelFormat := pf24bit;
              LineSize := Png.Width * 3;

              for Row := 0 to Png.Height -1 do
                Move (Png.Scanline[Row]^, Bmp.Scanline[Row]^, LineSize);
            end;
          COLOR_RGBALPHA:
            begin
              Bmp.PixelFormat := pf32bit;

              for Row := 0 to Png.Height -1 do begin
                pSource := Png.Scanline[Row];
                pDest := Bmp.Scanline[Row];

                for Col := 0 to Png.Width -1 do begin
                  Move (pSource^, pDest^, 3);
                  Inc(pSource, 3);
                  Inc(pDest, 3);

                  pDest^ := Png.AlphaScanline[Row][Col];
                  Inc(pDest);
                end;
              end;
            end;
          else
            Raise EglBitmapException.Create ('TglBitmap.LoadPng - Unsupported Colortype found.');
        end;

        Result := AssignFromBitmap (Bmp);
      finally
        Bmp.Free;
      end;
    finally
      Png.Free;
    end;
  end;
end;
{$endif}


procedure glBitmapSwapColors(Sender : TglBitmap; const Position, Size: TglBitmapPixelPosition; const Source, Dest: TglBitmapPixelData; const Data: Pointer);
var
  Temp: Byte;
begin
  Temp := Dest.ptRed^;
  Dest.ptRed^  := Dest.ptBlue^;
  Dest.ptBlue^ := Temp;
end;


procedure TglBitmap.SwapColors;
begin
  AddFunc(glBitmapSwapColors, False);
end;


procedure TglBitmap.Bind(EnableTextureUnit: Boolean);
begin
  if EnableTextureUnit
    then glEnable(TextureTarget);

  if FTextureID > 0
    then glBindTexture(TextureTarget, FTextureID);
end;


procedure TglBitmap.GetPixel(const Pos: TglBitmapPixelPosition;
  var Pixel: TglBitmapPixelData);
begin
  if Assigned (FGetPixelFunc)
    then FGetPixelFunc(Pos, Pixel);
end;


procedure TglBitmap.CreateTextureID;
begin
  // Generate Texture
  if TextureID <> 0
    then glDeleteTextures(1, @TextureID);

  glGenTextures(1, @FTextureID);

  glBindTexture(TextureTarget, TextureID);
end;


procedure TglBitmap.SetupTextureParameters(var BuildWithGlu: Boolean);
begin
  // Set up parameters
  glTexParameteri(TextureTarget, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(TextureTarget, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(TextureTarget, GL_TEXTURE_WRAP_R, GL_REPEAT);

  glTexParameteri(TextureTarget, GL_TEXTURE_MAG_FILTER, FFilterMag);
  if (MipMap = mmNone) then begin
    case FFilterMin of
      GL_NEAREST, GL_LINEAR:
        glTexParameteri(TextureTarget, GL_TEXTURE_MIN_FILTER, FFilterMin);
      GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST_MIPMAP_LINEAR:
        glTexParameteri(TextureTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_LINEAR:
        glTexParameteri(TextureTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    end;
  end else
    glTexParameteri(TextureTarget, GL_TEXTURE_MIN_FILTER, FFilterMin);

  // Mip Maps generation Mode
  BuildWithGlu := False;

  if (MipMap = mmMipmap) then begin
    if (GL_VERSION_1_4 or GL_SGIS_generate_mipmap)
      then glTexParameteri(TextureTarget, GL_GENERATE_MIPMAP, GL_TRUE)
      else BuildWithGlu := True;
  end else
  if (MipMap = mmMipmapGlu)
    then BuildWithGlu := True;
end;


procedure TglBitmap.SelectTextureFormat(var Format, Components: Cardinal);
begin
  // selecting Format
  if ((not GL_VERSION_1_2) and (not GL_EXT_bgra))
    then SwapColors;

  if (HasAlpha) then begin
    if (GL_VERSION_1_2) then begin
      Format := GL_BGRA;
    end else begin
      if (not GL_EXT_bgra)
        then Format := GL_RGBA
        else Format := GL_BGRA_EXT;
    end;
  end else begin
    if (GL_VERSION_1_2) then begin
      Format := GL_BGR;
    end else begin
      if (not GL_EXT_bgra)
        then Format := GL_RGB
        else Format := GL_BGR_EXT;
    end;
  end;

  // Selecting Components
  if (HasAlpha) then begin
    case FTextureFormat of
      tf16Bit:
        Components := GL_RGBA4;
      tf32Bit:
        Components := GL_RGBA8;
      tfCompressed:
        begin
          if (GL_ARB_texture_compression) then begin
            Components := GL_COMPRESSED_RGBA_ARB
          end else begin
            if (GL_EXT_texture_compression_s3tc)
              then Components := GL_COMPRESSED_RGBA_S3TC_DXT1_EXT
              else Components := GL_RGBA;
          end;
        end;
    else
      Components := GL_RGBA;
    end;
  end else begin
    case FTextureFormat of
      tf16Bit:
        Components := GL_RGB4;
      tf32Bit:
        Components := GL_RGB8;
      tfCompressed:
        begin
          if (GL_ARB_texture_compression) then begin
            Components := GL_COMPRESSED_RGB_ARB
          end else begin
            if (GL_EXT_texture_compression_s3tc)
              then Components := GL_COMPRESSED_RGB_S3TC_DXT1_EXT
              else Components := GL_RGB;
          end;
        end;
    else
      Components := GL_RGB;
    end;
  end;
end;


{ TglBitmap2D }

constructor TglBitmap2D.Create(Width, Height: Integer; Func: TglBitmapFunction;
  HasAlpha: Boolean);
begin
  inherited Create();

  LoadFromFunc(Width, Height, Func, HasAlpha);
end;


procedure TglBitmap2D.SetDataPtr(Ptr: PByte);
var
  Idx, LineWidth: Integer;
begin
  inherited;

  if HasAlpha
    then FGetPixelFunc := GetPixel2DAlpha
    else FGetPixelFunc := GetPixel2D;

  if Assigned(GetData()) then begin
    SetLength(FLines, FHeight);

    if HasAlpha
      then LineWidth := FWidth * 4
      else LineWidth := FWidth * 3;

    for Idx := 0 to FHeight -1
      do FLines [Idx] := PByte(Integer(GetData) + (Idx * LineWidth));
  end
    else SetLength(FLines, 0);
end;


procedure TglBitmap2D.LoadFromFunc(Width, Height: Integer;
  Func: TglBitmapFunction; HasAlpha: Boolean);
var
  Size: Integer;
begin
  if (HasAlpha)
    then Size := 4
    else Size := 3;

  FHasAlpha := HasAlpha;
  FWidth := Width;
  FHeight := Height;

  AllocData(Width * Height * Size);

  AddFunc(Func, False);
end;


function TglBitmap2D.IntAddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
var
  Col, Row: Integer;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
  pTempImage: PByte;
begin
  Result := False;

  pTempImage := nil;

  // empty records
  ZeroMemory (@Pos,    SizeOf (Pos));
  ZeroMemory (@Size,   SizeOf (Size));
  ZeroMemory (@Dest,   SizeOf (Dest));
  ZeroMemory (@Source, SizeOf (Source));

  // Prepare
  Size.X := Width;
  Size.Y := Height;
  Size.Fields := [ffX, ffY];

  Pos.Fields := [ffX, ffY];

  Source.Fields  := [ffRed, ffGreen, ffBlue];
  Source.ptRed   := GetData();
  Source.ptGreen := GetData();
  Source.ptBlue  := GetData();
  Inc(Source.ptRed,   2);
  Inc(Source.ptGreen, 1);

  if CreateTemp then begin
    pTempImage := AllocMem (Width * Height * 3);

    Dest.Fields  := Source.Fields;
    Dest.ptRed   := pTempImage;
    Dest.ptGreen := pTempImage;
    Dest.ptBlue  := pTempImage;
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    Move (Source.ptBlue^, pTempImage^, Width * Height * 3);
  end else begin
    Dest.Fields  := Source.Fields;
    Dest.ptRed   := Source.ptRed;
    Dest.ptGreen := Source.ptGreen;
    Dest.ptBlue  := Source.ptBlue;
  end;

  for Row := 0 to Height -1 do begin
    Pos.Y := Row;

    for Col := 0 to Width -1 do begin
      Pos.X := Col;
      Func (Self, Pos, Size, Source, Dest, Data);

      Inc(Source.ptRed,   3);
      Inc(Source.ptGreen, 3);
      Inc(Source.ptBlue,  3);

      Inc(Dest.ptRed,   3);
      Inc(Dest.ptGreen, 3);
      Inc(Dest.ptBlue,  3);
    end;
  end;

  if CreateTemp
    then SetDataPtr(pTempImage);
end;


function TglBitmap2D.IntAddFuncAlpha(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
var
  Col, Row: Integer;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
  pTempImage: PByte;
begin
  Result := False;

  pTempImage := nil;

  // empty records
  ZeroMemory (@Pos,    SizeOf (Pos));
  ZeroMemory (@Size,   SizeOf (Size));
  ZeroMemory (@Dest,   SizeOf (Dest));
  ZeroMemory (@Source, SizeOf (Source));

  // Prepare
  Size.X := Width;
  Size.Y := Height;
  Size.Fields := [ffX, ffY];

  Pos.Fields := [ffX, ffY];

  Source.Fields  := [ffRed, ffGreen, ffBlue, ffAlpha];
  Source.ptRed   := GetData();
  Source.ptGreen := GetData();
  Source.ptBlue  := GetData();
  Source.ptAlpha := GetData();
  Inc(Source.ptAlpha, 3);
  Inc(Source.ptRed,   2);
  Inc(Source.ptGreen, 1);

  if CreateTemp then begin
    pTempImage := AllocMem (Width * Height * 4);

    Dest.Fields  := Source.Fields;
    Dest.ptRed   := pTempImage;
    Dest.ptGreen := pTempImage;
    Dest.ptBlue  := pTempImage;
    Dest.ptAlpha := pTempImage;
    Inc(Dest.ptAlpha, 3);
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    Move (Source.ptBlue^, pTempImage^, Width * Height * 3);
  end else begin
    Dest.Fields  := Source.Fields;
    Dest.ptRed   := Source.ptRed;
    Dest.ptGreen := Source.ptGreen;
    Dest.ptBlue  := Source.ptBlue;
    Dest.ptAlpha := Source.ptAlpha;
  end;

  for Row := 0 to Height -1 do begin
    Pos.Y := Row;

    for Col := 0 to Width -1 do begin
      Pos.X := Col;
      Func (Self, Pos, Size, Source, Dest, Data);

      Inc(Source.ptRed,   4);
      Inc(Source.ptGreen, 4);
      Inc(Source.ptBlue,  4);
      Inc(Source.ptAlpha, 4);

      Inc(Dest.ptRed,   4);
      Inc(Dest.ptGreen, 4);
      Inc(Dest.ptBlue,  4);
      Inc(Dest.ptAlpha, 4);
    end;
  end;

  if CreateTemp
    then SetDataPtr(pTempImage);
end;


function TglBitmap2D.AddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
begin
  assert (Assigned (GetData()));

  if HasAlpha
    then Result := IntAddFuncAlpha(Func, CreateTemp, Data)
    else Result := IntAddFunc(Func, CreateTemp, Data);
end;


function TglBitmap2D.AssignToBitmap(const Bitmap: TBitmap): boolean;
var
  Row, Size, RowSize: Integer;
  pSource, pData: PByte;
begin
  Result := False;

  if Assigned(GetData()) then begin
    if Assigned(Bitmap) then begin
      Bitmap.Width := Width;
      Bitmap.Height := Height;
      Bitmap.PixelFormat := pf24bit;

      // Copy Data
      pSource := GetData();

      if HasAlpha then begin
        Size := 4;
        Bitmap.PixelFormat := pf32bit;
      end else begin
        Size := 3;
        Bitmap.PixelFormat := pf24bit;
      end;

      RowSize := Width * Size;

      for Row := 0 to Height -1 do begin
        pData := Bitmap.Scanline[Row];
        if Assigned(pData) then begin
          Move(pSource^, pData^, RowSize);
          Inc(pSource, RowSize);
        end;
      end;

      Result := True;
    end;
  end;
end;


function TglBitmap2D.AssignAlphaToBitmap(const Bitmap: TBitmap): boolean;
var
  Row, Col: Integer;
  pSource, pDest: PByte;
begin
  Result := False;

  if Assigned(GetData()) then begin
    if (HasAlpha) then begin
      if Assigned(Bitmap) then begin
        Bitmap.Width := Width;
        Bitmap.Height := Height;
        Bitmap.PixelFormat := pf24bit;

        // Copy Data
        pSource := GetData();

        for Row := 0 to Height -1 do begin
          pDest := Bitmap.Scanline[Row];
          if Assigned(pDest) then begin
            for Col := 0 to Width -1 do begin
              Inc(pSource, 3);
              pDest^ := pSource^;
              Inc(pDest, 1);
              pDest^ := pSource^;
              Inc(pDest, 1);
              pDest^ := pSource^;
              Inc(pDest, 1);
              Inc(pSource, 1);
            end;
          end;
        end;

        Result := True;
      end;
    end;
  end;
end;


function TglBitmap2D.AssignFromBitmap(const Bitmap: TBitmap): boolean;
var
  pSource, pData: PByte;
  Row, Size, RowSize: Integer;
begin
  Result := False;

  if (Assigned(Bitmap)) then begin
    if ((Bitmap.PixelFormat <> pf24Bit) and (Bitmap.PixelFormat <> pf32Bit))
      then raise EglBitmapException.Create('TglBitmap2D.AssignFromBitmap - Only Bitmaps with 24 or 32 bit are Supported. Set the pixelformat.');

    // Copy Data
    if (Bitmap.PixelFormat = pf24Bit) then begin
      FHasAlpha := False;
      Size := 3;
    end else begin
      FHasAlpha := True;
      Size := 4;
    end;
    FWidth := Bitmap.Width;
    FHeight := Bitmap.Height;

    RowSize := FWidth * Size;

    AllocData(Height * RowSize);
    pData := GetData;

    for Row := 0 to Height -1 do begin
      pSource := Bitmap.Scanline[Row];

      if (Assigned(pSource)) then begin
        Move(pSource^, pData^, RowSize);
        Inc(pData, RowSize);
      end;
    end;
    Result := True;
  end;
end;


function TglBitmap2D.AddAlphaFromFunc(
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  Col, Row, TempSize, PixSize: Integer;
  pNewImage, pSource, pDest: pByte;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
begin
  Result := false;

  if (Assigned(GetData())) then begin
    TempSize := Width * Height * 4;
    pNewImage := AllocMem(TempSize);
    pDest := pNewImage;

    // empty records
    ZeroMemory (@Pos,    SizeOf (Pos));
    ZeroMemory (@Size,   SizeOf (Size));
    ZeroMemory (@Dest,   SizeOf (Dest));
    ZeroMemory (@Source, SizeOf (Source));

    // Prepare
    Size.X := Width;
    Size.Y := Height;
    Size.Fields := [ffX, ffY];

    Pos.Fields := [ffX, ffY];

    Source.Fields := [ffRed, ffGreen, ffBlue];
    Source.ptRed   := GetData();
    Source.ptGreen := GetData();
    Source.ptBlue  := GetData();
    Inc(Source.ptRed,   2);
    Inc(Source.ptGreen, 1);

    pSource := GetData;
    if HasAlpha then begin
      Source.ptAlpha := GetData();
      Inc(Source.ptAlpha, 3);
      PixSize := 4;
    end else begin
      PixSize := 3;
    end;

    Dest.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
    Dest.ptRed   := pDest;
    Dest.ptGreen := pDest;
    Dest.ptBlue  := pDest;
    Dest.ptAlpha := pDest;
    Inc(Dest.ptAlpha, 3);
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    // Copy Pixels
    for Row := 0 to Height -1 do begin
      Pos.Y := Row;

      for Col := 0 to Width -1 do begin
        Pos.X := Col;

        MoveMemory(pDest, pSource, PixSize);
        Inc(pSource, PixSize);
        Inc(pDest,   4);

        Func (Self, Pos, Size, Source, Dest, Data);

        Inc(Source.ptRed,   PixSize);
        Inc(Source.ptGreen, PixSize);
        Inc(Source.ptBlue,  PixSize);
        if HasAlpha then
          Inc(Source.ptAlpha, PixSize);

        Inc(Dest.ptRed,   4);
        Inc(Dest.ptGreen, 4);
        Inc(Dest.ptBlue,  4);
        Inc(Dest.ptAlpha, 4);
      end;
    end;

    FHasAlpha := True;
    SetDataPtr(pNewImage);
    Result := True;
  end;
end;


function TglBitmap2D.AddAlphaFromBitmap(Bitmap: TBitmap;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  Col, Row, TempSize, PixSize: Integer;
  pNewImage, pSource, pDest: pByte;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
begin
  Result := False;
  assert(Bitmap.PixelFormat = pf24Bit, 'TglBitmap2D.AddAlphaFromBitmap - Only 24Bit Bitmaps supported.');

  if ((Bitmap.Width = Width) and (Bitmap.Height = Height)) then begin
    if (Assigned(GetData())) then begin
      TempSize := Width * Height * 4;
      pNewImage := AllocMem(TempSize);
      try
        pDest := pNewImage;

        if not Assigned(Func)
          then Func := glBitmapDefaultAlphaFunc;

        // empty records
        ZeroMemory (@Pos,    SizeOf (Pos));
        ZeroMemory (@Size,   SizeOf (Size));
        ZeroMemory (@Dest,   SizeOf (Dest));
        ZeroMemory (@Source, SizeOf (Source));

        // Prepare
        Size.X := Width;
        Size.Y := Height;
        Size.Fields := [ffX, ffY];

        Pos.Fields := [ffX, ffY];

        Source.Fields := [ffRed, ffGreen, ffBlue];

        Dest.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
        Dest.ptRed   := pDest;
        Dest.ptGreen := pDest;
        Dest.ptBlue  := pDest;
        Dest.ptAlpha := pDest;
        Inc(Dest.ptAlpha, 3);
        Inc(Dest.ptRed,   2);
        Inc(Dest.ptGreen, 1);

        pSource := GetData;
        if (HasAlpha)
          then PixSize := 4
          else PixSize := 3;

        // Copy Pixels
        for Row := 0 to Height -1 do begin
          Pos.Y := Row;

          Source.ptRed   := Bitmap.ScanLine[Row];
          Source.ptGreen := Bitmap.ScanLine[Row];
          Source.ptBlue  := Bitmap.ScanLine[Row];
          Inc(Source.ptRed, 2);
          Inc(Source.ptGreen, 1);

          for Col := 0 to Width -1 do begin
            Pos.X := Col;

            MoveMemory(pDest, pSource, PixSize);
            Inc(pSource, PixSize);
            Inc(pDest,   4);

            Func (Self, Pos, Size, Source, Dest, Data);

            Inc(Dest.ptRed,   4);
            Inc(Dest.ptGreen, 4);
            Inc(Dest.ptBlue,  4);
            Inc(Dest.ptAlpha, 4);

            Inc(Source.ptRed,   3);
            Inc(Source.ptGreen, 3);
            Inc(Source.ptBlue,  3);
          end;
        end;

        // If we have Alpha we don't need to realloc
        if not FHasAlpha then begin
          TempSize := Width * Height * 4;
          AllocData(TempSize);
        end;

        // Copy image to Original
        pDest := GetData;
        MoveMemory(pDest, pNewImage, TempSize);

        FHasAlpha := True;
      finally
        FreeMem(pNewImage);
      end;
    end;
  end;
end;


function TglBitmap2D.RemoveAlpha: Boolean;
var
  pSource, pDest, pTempDest: PByte;
  Row, Col, Size: Integer;
begin
  Result := False;

  if (Assigned(Getdata())) then begin
    if (FHasAlpha) then begin
      Size := Width * Height * 3;
      pSource := GetData;
      pDest := AllocMem(Size);
      pTempDest := pDest;

      for Row := 0 to Height -1 do begin
        for Col := 0 to Width -1 do begin
          Move(pSource^, pTempDest^, 3);
          inc(pSource, 4);
          inc(pTempDest, 3);
        end;
      end;

      FHasAlpha := False;
      SetDataPtr(pDest);
    end;
  end;
end;


procedure TglBitmap2D.GetPixel2D(const Pos: TglBitmapPixelPosition;
  var Pixel: TglBitmapPixelData);
var
  BasePtr: Integer;
begin
  inherited;

  if (Pos.Y <= Height) and (Pos.X <= Width) then
    begin
      BasePtr := Integer(FLines[Pos.Y]) + Pos.X * 3;

      Pixel.Fields := [ffRed, ffGreen, ffBlue];
      Pixel.ptRed   := PByte(BasePtr + 2);
      Pixel.ptGreen := PByte(BasePtr + 1);
      Pixel.ptBlue  := PByte(BasePtr);
    end;
end;


procedure TglBitmap2D.GetPixel2DAlpha(const Pos: TglBitmapPixelPosition;
  var Pixel: TglBitmapPixelData);
var
  BasePtr: Integer;
begin
  inherited;

  if (Pos.Y <= Height) and (Pos.X <= Width) then
    begin
      BasePtr := Integer(FLines[Pos.Y]) + Pos.X * 4;

      Pixel.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
      Pixel.ptAlpha := PByte(BasePtr + 3);
      Pixel.ptRed   := PByte(BasePtr + 2);
      Pixel.ptGreen := PByte(BasePtr + 1);
      Pixel.ptBlue  := PByte(BasePtr);
    end;
end;


function TglBitmap2D.FlipHorz: Boolean;
var
  Col, Row: Integer;
  pTempDest, pDest, pSource: pByte;
  Size, RowSize, ImgSize: Integer;
begin
  Result := False;

  if Assigned(GetData()) then begin
    pSource := GetData();
    if (HasAlpha)
      then Size := 4
      else Size := 3;

    RowSize := Width * Size;
    ImgSize := Height * RowSize;

    pDest := AllocMem(ImgSize);
    pTempDest := pDest;

    Dec(pTempDest, RowSize + Size);
    for Row := 0 to Height -1 do begin
      Inc(pTempDest, RowSize * 2);
      for Col := 0 to Width -1 do begin
        Move(pSource^, pTempDest^, Size);

        Inc(pSource, Size);
        Dec(pTempDest, Size);
      end;
    end;

    SetDataPtr(pDest);

    Result := True;
  end;
end;


function TglBitmap2D.FlipVert: Boolean;
var
  Row: Integer;
  pTempDest, pDest, pSource: pByte;
  Size, RowSize: Integer;
begin
  Result := False;

  if Assigned(GetData()) then begin
    pSource := GetData();
    if (HasAlpha)
      then Size := 4
      else Size := 3;

    RowSize := Width * Size;

    pDest := AllocMem(Height * RowSize);
    pTempDest := pDest;

    Inc(pTempDest, Width * (Height -1) * Size);

    for Row := 0 to Height -1 do begin
      Move(pSource^, pTempDest^, RowSize);

      Dec(pTempDest, RowSize);
      Inc(pSource, RowSize);
    end;

    SetDataPtr(pDest);

    Result := True;
  end;
end;


procedure TglBitmap2D.UploadTextureData (TextureTarget, Format, Components: Cardinal; BuildWithGlu: Boolean);
begin
  // Upload data
  if BuildWithGlu
    then gluBuild2DMipmaps(TextureTarget, Components, Width, Height, Format, GL_UNSIGNED_BYTE, PByte(GetData()))
    else glTexImage2D(TextureTarget, 0, Components, Width, Height, 0, Format, GL_UNSIGNED_BYTE, PByte(GetData()));

  // Freigeben
  if (FreeDataAfterGenTexture)
    then SetDataPtr(nil);
end;


procedure TglBitmap2D.GenTexture(TestTextureSize: Boolean);
var
  BuildWithGlu: Boolean;
  Format, Components: Cardinal;
  TexSize: TGLint;
begin
  if Assigned(GetData()) then begin
    // Check Texture Size
    if (TestTextureSize) then begin
      glGetIntegerv(GL_MAX_TEXTURE_SIZE, @TexSize);

      if ((Height > TexSize) or (Width > TexSize))
        then raise EglBitmapSizeToLargeException.Create('TglBitmap2D.GenTexture - The size for the texture is to large. It''s may be not conform with the Hardware.');
    end;

    CreateTextureId;

    SetupTextureParameters(BuildWithGlu);
    SelectTextureFormat(Format, Components);

    UploadTextureData(TextureTarget, Format, Components, BuildWithGlu);

    // Infos sammeln
    if (GL_ARB_texture_compression) then begin
      glGetTexLevelParameteriv(TextureTarget, 0, GL_TEXTURE_COMPRESSED_ARB, @TexSize);
      FIsCompressed := TexSize <> 0;
      glGetTexLevelParameteriv(TextureTarget, 0, GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB, @FTextureSize);
    end else begin
      FIsCompressed := ((FTextureFormat = tfCompressed) and (GL_EXT_texture_compression_s3tc));
      FTextureSize := 0;
    end;
    glAreTexturesResident(1, @TextureID, @FIsResident);
  end;
end;


procedure TglBitmap2D.AfterConstruction;
begin
  inherited;

  TextureTarget := GL_TEXTURE_2D;
  FGetPixelFunc := GetPixel2D;
end;


{ TglBitmap1D }

constructor TglBitmap1D.Create(Width: Integer; Func: TglBitmapFunction;
  HasAlpha: Boolean);
begin
  inherited Create();

  LoadFromFunc(Width, Func, HasAlpha);
end;


procedure TglBitmap1D.SetDataPtr(Ptr: PByte);
begin
  inherited;

  if HasAlpha
    then FGetPixelFunc := GetPixel1DAlpha
    else FGetPixelFunc := GetPixel1D;
end;


procedure TglBitmap1D.LoadFromFunc(Width: Integer;
  Func: TglBitmapFunction; HasAlpha: Boolean);
var
  Size: Integer;
begin
  if (HasAlpha)
    then Size := 4
    else Size := 3;

  FHasAlpha := HasAlpha;
  FWidth := Width;

  AllocData(Width * Size);

  AddFunc(Func, False);
end;


function TglBitmap1D.IntAddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
var
  Col: Integer;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
  pTempImage: PByte;
begin
  Result := False;

  pTempImage := nil;

  // empty records
  ZeroMemory (@Pos,    SizeOf (Pos));
  ZeroMemory (@Size,   SizeOf (Size));
  ZeroMemory (@Dest,   SizeOf (Dest));
  ZeroMemory (@Source, SizeOf (Source));

  // Prepare
  Size.X := Width;
  Size.Fields := [ffX];

  Pos.Fields := [ffX];

  Source.Fields  := [ffRed, ffGreen, ffBlue];
  Source.ptRed   := GetData();
  Source.ptGreen := GetData();
  Source.ptBlue  := GetData();
  Inc(Source.ptRed,   2);
  Inc(Source.ptGreen, 1);

  if CreateTemp then begin
    pTempImage := AllocMem (Width * 3);

    Dest.Fields  := Source.Fields;
    Dest.ptRed   := pTempImage;
    Dest.ptGreen := pTempImage;
    Dest.ptBlue  := pTempImage;
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    Move (Source.ptBlue^, pTempImage^, Width * 3);
  end else begin
    Dest.Fields  := Source.Fields;
    Dest.ptRed   := Source.ptRed;
    Dest.ptGreen := Source.ptGreen;
    Dest.ptBlue  := Source.ptBlue;
  end;

  for Col := 0 to Width -1 do begin
    Pos.X := Col;
    Func (Self, Pos, Size, Source, Dest, Data);

    Inc(Source.ptRed,   3);
    Inc(Source.ptGreen, 3);
    Inc(Source.ptBlue,  3);

    Inc(Dest.ptRed,   3);
    Inc(Dest.ptGreen, 3);
    Inc(Dest.ptBlue,  3);
  end;

  if CreateTemp
    then SetDataPtr(pTempImage);
end;


function TglBitmap1D.IntAddFuncAlpha(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
var
  Col: Integer;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
  pTempImage: PByte;
begin
  Result := False;

  pTempImage := nil;

  // empty records
  ZeroMemory (@Pos,    SizeOf (Pos));
  ZeroMemory (@Size,   SizeOf (Size));
  ZeroMemory (@Dest,   SizeOf (Dest));
  ZeroMemory (@Source, SizeOf (Source));

  // Prepare
  Size.X := Width;
  Size.Fields := [ffX];

  Pos.Fields := [ffX];

  Source.Fields  := [ffRed, ffGreen, ffBlue, ffAlpha];
  Source.ptRed   := GetData();
  Source.ptGreen := GetData();
  Source.ptBlue  := GetData();
  Source.ptAlpha := GetData();
  Inc(Source.ptAlpha, 3);
  Inc(Source.ptRed,   2);
  Inc(Source.ptGreen, 1);

  if CreateTemp then begin
    pTempImage := AllocMem (Width * 4);

    Dest.Fields  := Source.Fields;
    Dest.ptRed   := pTempImage;
    Dest.ptGreen := pTempImage;
    Dest.ptBlue  := pTempImage;
    Dest.ptAlpha := pTempImage;
    Inc(Dest.ptAlpha, 3);
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    Move (Source.ptBlue^, pTempImage^, Width * 3);
    // TODO: Was ist denn das?
  end else begin
    Dest.Fields  := Source.Fields;
    Dest.ptRed   := Source.ptRed;
    Dest.ptGreen := Source.ptGreen;
    Dest.ptBlue  := Source.ptBlue;
    Dest.ptAlpha := Source.ptAlpha;
  end;

  for Col := 0 to Width -1 do begin
    Pos.X := Col;
    Func (Self, Pos, Size, Source, Dest, Data);

    Inc(Source.ptRed,   4);
    Inc(Source.ptGreen, 4);
    Inc(Source.ptBlue,  4);
    Inc(Source.ptAlpha, 4);

    Inc(Dest.ptRed,   4);
    Inc(Dest.ptGreen, 4);
    Inc(Dest.ptBlue,  4);
    Inc(Dest.ptAlpha, 4);
  end;

  if CreateTemp
    then SetDataPtr(pTempImage);
end;


function TglBitmap1D.AddFunc(Func: TglBitmapFunction; CreateTemp: Boolean; Data: Pointer): boolean;
begin
  assert (Assigned (GetData()));

  if HasAlpha
    then Result := IntAddFuncAlpha(Func, CreateTemp, Data)
    else Result := IntAddFunc(Func, CreateTemp, Data);
end;


function TglBitmap1D.AssignToBitmap(const Bitmap: TBitmap): boolean;
var
  Size, RowSize: Integer;
  pSource, pData: PByte;
begin
  Result := False;

  if Assigned(GetData()) then begin
    if Assigned(Bitmap) then begin
      Bitmap.Width := Width;
      Bitmap.Height := 1;
      Bitmap.PixelFormat := pf24bit;

      // Copy Data
      pSource := GetData();

      if HasAlpha then begin
        Size := 4;
        Bitmap.PixelFormat := pf32bit;
      end else begin
        Size := 3;
        Bitmap.PixelFormat := pf24bit;
      end;

      RowSize := Width * Size;

      pData := Bitmap.Scanline[0];
      if Assigned(pData) then
        Move(pSource^, pData^, RowSize);

      Result := True;
    end;
  end;
end;


function TglBitmap1D.AssignAlphaToBitmap(const Bitmap: TBitmap): boolean;
var
  Col: Integer;
  pSource, pDest: PByte;
begin
  Result := False;

  if Assigned(GetData()) then begin
    if (HasAlpha) then begin
      if Assigned(Bitmap) then begin
        Bitmap.Width := Width;
        Bitmap.Height := 1;
        Bitmap.PixelFormat := pf24bit;

        // Copy Data
        pSource := GetData();

        pDest := Bitmap.Scanline[0];
        if Assigned(pDest) then begin
          for Col := 0 to Width -1 do begin
            Inc(pSource, 3);
            pDest^ := pSource^;
            Inc(pDest, 1);
            pDest^ := pSource^;
            Inc(pDest, 1);
            pDest^ := pSource^;
            Inc(pDest, 1);
            Inc(pSource, 1);
          end;
        end;

        Result := True;
      end;
    end;
  end;
end;


function TglBitmap1D.AssignFromBitmap(const Bitmap: TBitmap): boolean;
var
  pSource, pData: PByte;
  Size, RowSize: Integer;
begin
  Result := False;

  if (Assigned(Bitmap)) then begin
    if ((Bitmap.PixelFormat <> pf24Bit) and (Bitmap.PixelFormat <> pf32Bit))
      then raise EglBitmapException.Create('TglBitmap2D.AssignFromBitmap - Only Bitmaps with 24 or 32 bit are Supported. Set the pixelformat.');

    // Copy Data
    if (Bitmap.PixelFormat = pf24Bit) then begin
      FHasAlpha := False;
      Size := 3;
    end else begin
      FHasAlpha := True;
      Size := 4;
    end;
    FWidth := Bitmap.Width;

    RowSize := FWidth * Size;

    AllocData(RowSize);
    pData := GetData;

    pSource := Bitmap.Scanline[0];

    if (Assigned(pSource)) then 
      Move(pSource^, pData^, RowSize);

    Result := True;
  end;
end;


function TglBitmap1D.AddAlphaFromFunc(
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  Col, TempSize, PixSize: Integer;
  pNewImage, pSource, pDest: pByte;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
begin
  Result := false;

  if (Assigned(GetData())) then begin
    TempSize := Width * 4;
    pNewImage := AllocMem(TempSize);
    pDest := pNewImage;

    // empty records
    ZeroMemory (@Pos,    SizeOf (Pos));
    ZeroMemory (@Size,   SizeOf (Size));
    ZeroMemory (@Dest,   SizeOf (Dest));
    ZeroMemory (@Source, SizeOf (Source));

    // Prepare
    Size.X := Width;
    Size.Fields := [ffX];

    Pos.Fields := [ffX];

    Source.Fields := [ffRed, ffGreen, ffBlue];
    Source.ptRed   := GetData();
    Source.ptGreen := GetData();
    Source.ptBlue  := GetData();
    Inc(Source.ptRed,   2);
    Inc(Source.ptGreen, 1);

    pSource := GetData;
    if HasAlpha then begin
      Source.ptAlpha := GetData();
      Inc(Source.ptAlpha, 3);
      PixSize := 4;
    end else begin
      PixSize := 3;
    end;

    Dest.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
    Dest.ptRed   := pDest;
    Dest.ptGreen := pDest;
    Dest.ptBlue  := pDest;
    Dest.ptAlpha := pDest;
    Inc(Dest.ptAlpha, 3);
    Inc(Dest.ptRed,   2);
    Inc(Dest.ptGreen, 1);

    // Copy Pixels
    for Col := 0 to Width -1 do begin
      Pos.X := Col;

      MoveMemory(pDest, pSource, PixSize);
      Inc(pSource, PixSize);
      Inc(pDest,   4);

      Func (Self, Pos, Size, Source, Dest, Data);

      Inc(Source.ptRed,   PixSize);
      Inc(Source.ptGreen, PixSize);
      Inc(Source.ptBlue,  PixSize);
      if HasAlpha then
        Inc(Source.ptAlpha, PixSize);

      Inc(Dest.ptRed,   4);
      Inc(Dest.ptGreen, 4);
      Inc(Dest.ptBlue,  4);
      Inc(Dest.ptAlpha, 4);
    end;

    FHasAlpha := True;
    SetDataPtr(pNewImage);
    Result := True;
  end;
end;


function TglBitmap1D.AddAlphaFromBitmap(Bitmap: TBitmap;
  Func: TglBitmapFunction; Data: Pointer): boolean;
var
  Col, TempSize, PixSize: Integer;
  pNewImage, pSource, pDest: pByte;
  Pos, Size: TglBitmapPixelPosition;
  Source, Dest: TglBitmapPixelData;
begin
  Result := False;
  assert(Bitmap.PixelFormat = pf24Bit, 'TglBitmap1D.AddAlphaFromBitmap - Only 24Bit Bitmaps supported.');

  if (Bitmap.Width = Width) then begin
    if (Assigned(GetData())) then begin
      TempSize := Width * 4;
      pNewImage := AllocMem(TempSize);
      try
        pDest := pNewImage;

        if not Assigned(Func)
          then Func := glBitmapDefaultAlphaFunc;

        // empty records
        ZeroMemory (@Pos,    SizeOf (Pos));
        ZeroMemory (@Size,   SizeOf (Size));
        ZeroMemory (@Dest,   SizeOf (Dest));
        ZeroMemory (@Source, SizeOf (Source));

        // Prepare
        Size.X := Width;
        Size.Fields := [ffX];

        Pos.Fields := [ffX];

        Source.Fields := [ffRed, ffGreen, ffBlue];

        Dest.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
        Dest.ptRed   := pDest;
        Dest.ptGreen := pDest;
        Dest.ptBlue  := pDest;
        Dest.ptAlpha := pDest;
        Inc(Dest.ptAlpha, 3);
        Inc(Dest.ptRed,   2);
        Inc(Dest.ptGreen, 1);

        pSource := GetData;
        if (HasAlpha)
          then PixSize := 4
          else PixSize := 3;

        // Copy Pixels
        Source.ptRed   := Bitmap.ScanLine[0];
        Source.ptGreen := Bitmap.ScanLine[0];
        Source.ptBlue  := Bitmap.ScanLine[0];
        Inc(Source.ptRed, 2);
        Inc(Source.ptGreen, 1);

        for Col := 0 to Width -1 do begin
          Pos.X := Col;

          MoveMemory(pDest, pSource, PixSize);
          Inc(pSource, PixSize);
          Inc(pDest,   4);

          Func (Self, Pos, Size, Source, Dest, Data);

          Inc(Dest.ptRed,   4);
          Inc(Dest.ptGreen, 4);
          Inc(Dest.ptBlue,  4);
          Inc(Dest.ptAlpha, 4);

          Inc(Source.ptRed,   3);
          Inc(Source.ptGreen, 3);
          Inc(Source.ptBlue,  3);
        end;

        // If we have Alpha we don't need to realloc
        if not FHasAlpha then begin
          TempSize := Width * 4;
          AllocData(TempSize);
        end;

        // Copy image to Original
        pDest := GetData;
        MoveMemory(pDest, pNewImage, TempSize);

        FHasAlpha := True;
      finally
        FreeMem(pNewImage);
      end;
    end;
  end;
end;


function TglBitmap1D.RemoveAlpha: Boolean;
var
  pSource, pDest, pTempDest: PByte;
  Col, Size: Integer;
begin
  Result := False;

  if (Assigned(Getdata())) then begin
    if (FHasAlpha) then begin
      Size := Width * 3;
      pSource := GetData;
      pDest := AllocMem(Size);
      pTempDest := pDest;

      for Col := 0 to Width -1 do begin
        Move(pSource^, pTempDest^, 3);
        inc(pSource, 4);
        inc(pTempDest, 3);
      end;

      FHasAlpha := False;
      SetDataPtr(pDest);
    end;
  end;
end;


procedure TglBitmap1D.GetPixel1D(const Pos: TglBitmapPixelPosition;
  var Pixel: TglBitmapPixelData);
var
  BasePtr: Integer;
begin
  inherited;

  if Pos.X <= Width then
    begin
      BasePtr := Pos.X * 3;

      Pixel.Fields := [ffRed, ffGreen, ffBlue];
      Pixel.ptRed   := PByte(BasePtr + 2);
      Pixel.ptGreen := PByte(BasePtr + 1);
      Pixel.ptBlue  := PByte(BasePtr);
    end;
end;


procedure TglBitmap1D.GetPixel1DAlpha(const Pos: TglBitmapPixelPosition;
  var Pixel: TglBitmapPixelData);
var
  BasePtr: Integer;
begin
  inherited;

  if Pos.X <= Width then
    begin
      BasePtr := Pos.X * 4;

      Pixel.Fields := [ffRed, ffGreen, ffBlue, ffAlpha];
      Pixel.ptAlpha := PByte(BasePtr + 3);
      Pixel.ptRed   := PByte(BasePtr + 2);
      Pixel.ptGreen := PByte(BasePtr + 1);
      Pixel.ptBlue  := PByte(BasePtr);
    end;
end;


function TglBitmap1D.FlipHorz: Boolean;
var
  Col: Integer;
  pTempDest, pDest, pSource: pByte;
  Size, RowSize: Integer;
begin
  Result := False;

  if Assigned(GetData()) then begin
    pSource := GetData();
    if (HasAlpha)
      then Size := 4
      else Size := 3;

    RowSize := Width * Size;

    pDest := AllocMem(RowSize);
    pTempDest := pDest;

    Inc(pTempDest, RowSize);
    for Col := 0 to Width -1 do begin
      Move(pSource^, pTempDest^, Size);

      Inc(pSource, Size);
      Dec(pTempDest, Size);
    end;

    SetDataPtr(pDest);

    Result := True;
  end;
end;


procedure TglBitmap1D.UploadTextureData (TextureTarget, Format, Components: Cardinal; BuildWithGlu: Boolean);
begin
  // Upload data
  if BuildWithGlu
    then gluBuild1DMipmaps(TextureTarget, Components, Width, Format, GL_UNSIGNED_BYTE, PByte(GetData()))
    else glTexImage1D(TextureTarget, 0, Components, Width, 0, Format, GL_UNSIGNED_BYTE, PByte(GetData()));

  // Freigeben
  if (FreeDataAfterGenTexture)
    then SetDataPtr(nil);
end;


procedure TglBitmap1D.GenTexture(TestTextureSize: Boolean);
var
  BuildWithGlu: Boolean;
  Format, Components: Cardinal;
  TexSize: TGLint;
begin
  if Assigned(GetData()) then begin
    // Check Texture Size
    if (TestTextureSize) then begin
      glGetIntegerv(GL_MAX_TEXTURE_SIZE, @TexSize);

      if (Width > TexSize)
        then raise EglBitmapSizeToLargeException.Create('TglBitmap1D.GenTexture - The size for the texture is to large. It''s may be not conform with the Hardware.');
    end;

    CreateTextureId;

    SetupTextureParameters(BuildWithGlu);
    SelectTextureFormat(Format, Components);

    UploadTextureData(TextureTarget, Format, Components, BuildWithGlu);

    // Infos sammeln
    if (GL_ARB_texture_compression) then begin
      glGetTexLevelParameteriv(TextureTarget, 0, GL_TEXTURE_COMPRESSED_ARB, @TexSize);
      FIsCompressed := TexSize <> 0;
      glGetTexLevelParameteriv(TextureTarget, 0, GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB, @FTextureSize);
    end else begin
      FIsCompressed := ((FTextureFormat = tfCompressed) and (GL_EXT_texture_compression_s3tc));
      FTextureSize := 0;
    end;
    glAreTexturesResident(1, @TextureID, @FIsResident);
  end;
end;


procedure TglBitmap1D.AfterConstruction;
begin
  inherited;

  TextureTarget := GL_TEXTURE_1D;
  FGetPixelFunc := GetPixel1D;
end;


{ TglBitmapCubeMap }

procedure TglBitmapCubeMap.AfterConstruction;
begin
  inherited;

  if not (GL_VERSION_1_3 or GL_ARB_texture_cube_map or GL_EXT_texture_cube_map) then
    raise EglBitmapException.Create('TglBitmapCubeMap.AfterConstruction - CubeMaps are unsupported.');

  SetTextureWrap; // set all to GL_CLAMP_TO_EDGE
  TextureTarget := GL_TEXTURE_CUBE_MAP;
end;


procedure TglBitmapCubeMap.Bind(EnableTexCoordsGen, EnableTextureUnit: Boolean);
begin
  inherited Bind (EnableTextureUnit);

  if EnableTexCoordsGen then begin
    glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    glTexGenf(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
  end;
end;


procedure TglBitmapCubeMap.GenerateCubeMap(CubeTarget: Cardinal);
var
  Format, Components: Cardinal;
  BuildWithGlu: Boolean;
begin
  if MipMap = mmMipmapGlu then
    MipMap := mmMipmap;

  // create Texture
  if TextureID = 0 then begin
    CreateTextureID;
    SetupTextureParameters(BuildWithGlu);
  end;

  SelectTextureFormat(Format, Components);

  // Never use gluBuild2DMipmaps to UploadTextureData 
  UploadTextureData (CubeTarget, Format, Components, False);
end;


procedure TglBitmapCubeMap.GenTexture(TestTextureSize: Boolean);
begin
  Assert(False, 'TglBitmapCubeMap.GenTexture - Don''t call GenTextures directly.');
end;


initialization
  glBitmapSetDefaultTextureFormat(tfDefault);
  glBitmapSetDefaultFilter(GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
  glBitmapSetDefaultTextureWrap(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);

  glBitmapSetDefaultFreeDataAfterGenTexture(True);
  glBitmapSetDefaultDeleteTextureOnFree(True);

finalization

end.

